@@ -3,7 +3,7 @@
  * 
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  * Copyright (C) 2000 Andre Hedrick <andre@linux-ide.org>
  *
@@ -29,13 +29,13 @@
 #include <stdlib.h>
 #include <ctype.h>
 
-#include "atacmds.h"
 #include "config.h"
 #include "int64.h"
+#include "atacmds.h"
 #include "extern.h"
 #include "utility.h"
 
-const char *atacmds_c_cvsid="$Id: atacmds.c,v 1.161 2004/08/18 19:27:44 likewise Exp $"
+const char *atacmds_c_cvsid="$Id: atacmds.c,v 1.168 2006/04/12 17:01:46 ballen4705 Exp $"
 ATACMDS_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID UTILITY_H_CVSID;
 
 // to hold onto exit code for atexit routine
@@ -608,6 +608,8 @@ int smartcommandhandler(int device, smart_command_set command, int select, char
   case CONTROLLER_3WARE_678K_CHAR:
   case CONTROLLER_3WARE_9000_CHAR:
     retval=escalade_command_interface(device, con->controller_port-1, con->controller_type, command, select, data);
+    if (retval &&  con->controller_port<=0)
+      pout("WARNING: apparently missing '-d 3ware,N' disk specification\n");
     break;
   case CONTROLLER_MARVELL_SATA:
     retval=marvell_command_interface(device, command, select, data);
@@ -628,7 +630,7 @@ int smartcommandhandler(int device, smart_command_set command, int select, char
     // if requested, pretty-print the output data structure
     if (con->reportataioctl>1 && getsdata) {
       if (command==CHECK_POWER_MODE)
-	pout("Sector Count Register (BASE-16): %02x\n", *data);
+	pout("Sector Count Register (BASE-16): %02x\n", (unsigned char)(*data));
       else
 	prettyprint((unsigned char *)data, commandstrings[command]);
     }
@@ -689,7 +691,9 @@ int ataReadHDIdentity (int device, struct ata_identify_device *buf){
     }
   }
 
+#ifndef __NetBSD__
   // if machine is big-endian, swap byte order as needed
+  // (the NetBSD kernel does deliver the results in host byte order)
   if (isbigendian()){
     int i;
     
@@ -703,6 +707,7 @@ int ataReadHDIdentity (int device, struct ata_identify_device *buf){
     for (i=0; i<168; i++)
       swap2((char *)(buf->words088_255+i));
   }
+#endif
   
   // If there is a checksum there, validate it
   if ((rawshort[255] & 0x00ff) == 0x00a5 && checksum(rawbyte))
@@ -1173,10 +1178,16 @@ int ataDisableAutoOffline (int device ){
 }
 
 // If SMART is enabled, supported, and working, then this call is
-// guaranteed to return 1, else zero.  Silent inverse of
-// ataSmartStatus()
-int ataDoesSmartWork(int device){       
-  return !smartcommandhandler(device, STATUS, 0, NULL);
+// guaranteed to return 1, else zero.  Note that it should return 1
+// regardless of whether the disk's SMART status is 'healthy' or
+// 'failing'.
+int ataDoesSmartWork(int device){
+  int retval=smartcommandhandler(device, STATUS, 0, NULL);
+
+  if (-1 == retval)
+    return 0;
+
+  return 1;
 }
 
 // This function uses a different interface (DRIVE_TASK) than the
