@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -25,9 +25,23 @@
 #ifndef ATACMDS_H_
 #define ATACMDS_H_
 
-#define ATACMDS_H_CVSID "$Id: atacmds.h,v 1.73 2004/08/18 19:27:44 likewise Exp $\n"
+#define ATACMDS_H_CVSID "$Id: atacmds.h,v 1.81 2006/04/12 14:54:28 ballen4705 Exp $\n"
 
-#include "int64.h"
+// Macro to check expected size of struct at compile time using a
+// dummy typedef.  On size mismatch, compiler reports a negative array
+// size.  If you see an error message of this form, it means that the
+// #pragma pack(1) pragma below is not having the desired effect on
+// your compiler.
+#define ASSERT_SIZEOF_STRUCT(s, n) \
+  typedef char assert_sizeof_struct_##s[(sizeof(struct s) == (n)) ? 1 : -1]
+
+// Add __attribute__((packed)) if compiler supports it
+// because some gcc versions (at least ARM) lack support of #pragma pack()
+#ifdef HAVE_ATTR_PACKED
+#define ATTR_PACKED __attribute__((packed))
+#else
+#define ATTR_PACKED
+#endif
 
 typedef enum {
   // returns no data, just succeeds or fails
@@ -108,8 +122,9 @@ struct ata_identify_device {
   unsigned short word086;
   unsigned short csf_default;
   unsigned short words088_255[168];
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_identify_device, 512);
 
 /* ata_smart_attribute is the vendor specific in SFF-8035 spec */ 
 #pragma pack(1)
@@ -122,8 +137,9 @@ struct ata_smart_attribute {
   unsigned char worst;
   unsigned char raw[6];
   unsigned char reserv;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_attribute, 12);
 
 // MACROS to interpret the flags bits in the previous structure.
 // These have not been implemented using bitflags and a union, to make
@@ -189,8 +205,9 @@ struct ata_smart_values {
   unsigned char reserved_375_385[11];
   unsigned char vendor_specific_386_510[125]; // Maxtor bytes 508-509 Attribute/Threshold Revision #
   unsigned char chksum;
-}; 
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_values, 512);
 
 /* Maxtor, IBM: self-test failure checkpoint byte meaning:
  00 - write test
@@ -207,8 +224,9 @@ struct ata_smart_threshold_entry {
   unsigned char id;
   unsigned char threshold;
   unsigned char reserved[10];
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_threshold_entry, 12);
 
 /* Format of Read SMART THreshold Command */
 /* Compare to ata_smart_values above */
@@ -218,8 +236,9 @@ struct ata_smart_thresholds_pvt {
   struct ata_smart_threshold_entry thres_entries[NUMBER_ATA_SMART_ATTRIBUTES];
   unsigned char reserved[149];
   unsigned char chksum;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_thresholds_pvt, 512);
 
 
 // Table 42 of T13/1321D Rev 1 spec (Error Data Structure)
@@ -236,8 +255,9 @@ struct ata_smart_errorlog_error_struct {
   unsigned char extended_error[19];
   unsigned char state;
   unsigned short timestamp;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_errorlog_error_struct, 30);
 
 
 // Table 41 of T13/1321D Rev 1 spec (Command Data Structure)
@@ -252,16 +272,18 @@ struct ata_smart_errorlog_command_struct {
   unsigned char drive_head;
   unsigned char commandreg;
   unsigned int timestamp;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_errorlog_command_struct, 12);
 
 // Table 40 of T13/1321D Rev 1 spec (Error log data structure)
 #pragma pack(1)
 struct ata_smart_errorlog_struct {
   struct ata_smart_errorlog_command_struct commands[5];
   struct ata_smart_errorlog_error_struct error_struct;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_errorlog_struct, 90);
 
 // Table 39 of T13/1321D Rev 1 spec (SMART error log sector)
 #pragma pack(1)
@@ -272,8 +294,9 @@ struct ata_smart_errorlog {
   unsigned short int ata_error_count;
   unsigned char reserved[57];
   unsigned char checksum;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_errorlog, 512);
 
 // Table 45 of T13/1321D Rev 1 spec (Self-test log descriptor entry)
 #pragma pack(1)
@@ -284,8 +307,9 @@ struct ata_smart_selftestlog_struct {
   unsigned char selftestfailurecheckpoint;
   unsigned int lbafirstfailure;
   unsigned char vendorspecific[15];
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_selftestlog_struct, 24);
 
 // Table 44 of T13/1321D Rev 1 spec (Self-test log data structure)
 #pragma pack(1)
@@ -296,23 +320,26 @@ struct ata_smart_selftestlog {
   unsigned char mostrecenttest;
   unsigned char reserved[2];
   unsigned char chksum;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_selftestlog, 512);
 
 // SMART LOG DIRECTORY Table 52 of T13/1532D Vol 1 Rev 1a
 #pragma pack(1)
 struct ata_smart_log_entry {
   unsigned char numsectors;
   unsigned char reserved;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_log_entry, 2);
 
 #pragma pack(1)
 struct ata_smart_log_directory {
   unsigned short int logversion;
   struct ata_smart_log_entry entry[255];
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_smart_log_directory, 512);
 
 // SMART SELECTIVE SELF-TEST LOG Table 61 of T13/1532D Volume 1
 // Revision 3
@@ -320,8 +347,9 @@ struct ata_smart_log_directory {
 struct test_span {
   uint64_t start;
   uint64_t end;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(test_span, 16);
 
 #pragma pack(1)
 struct ata_selective_self_test_log {
@@ -336,8 +364,9 @@ struct ata_selective_self_test_log {
   unsigned short     pendingtime;
   unsigned char      reserved2;
   unsigned char      checksum;
-};
+} ATTR_PACKED;
 #pragma pack()
+ASSERT_SIZEOF_STRUCT(ata_selective_self_test_log, 512);
 
 #define SELECTIVE_FLAG_DOSCAN  (0x0002)
 #define SELECTIVE_FLAG_PENDING (0x0008)
@@ -492,6 +521,11 @@ char *create_vendor_attribute_arg_list(void);
 int ata_command_interface(int device, smart_command_set command, int select, char *data);
 int escalade_command_interface(int fd, int escalade_port, int escalade_type, smart_command_set command, int select, char *data);
 int marvell_command_interface(int device, smart_command_set command, int select, char *data);
+// Optional functions of os_*.c
+#ifdef HAVE_ATA_IDENTIFY_IS_CACHED
+// Return true if OS caches the ATA identify sector
+int ata_identify_is_cached(int fd);
+#endif
 
 // This function is exported to give low-level capability
 int smartcommandhandler(int device, smart_command_set command, int select, char *data);
