@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,20 +22,25 @@
  *
  */
 
+#include "config.h"
+
 #include <ctype.h>
 #include <stdio.h>
 #include <string.h>
+#ifdef HAVE_LOCALE_H
+#include <locale.h>
+#endif // #ifdef HAVE_LOCALE_H
+
+#include "int64.h"
 #include "atacmdnames.h"
 #include "atacmds.h"
 #include "ataprint.h"
 #include "smartctl.h"
-#include "int64.h"
 #include "extern.h"
 #include "utility.h"
 #include "knowndrives.h"
-#include "config.h"
 
-const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.156 2004/09/10 04:13:41 ballen4705 Exp $"
+const char *ataprint_c_cvsid="$Id: ataprint.c,v 1.164 2006/04/12 14:54:28 ballen4705 Exp $"
 ATACMDNAMES_H_CVSID ATACMDS_H_CVSID ATAPRINT_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID KNOWNDRIVES_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // for passing global control variables
@@ -90,22 +95,23 @@ void formatdriveidstring(char *out, const char *in, int n)
   char tmp[65];
 
   n = n > 64 ? 64 : n;
+#ifndef __NetBSD__
   swapbytes(tmp, in, n);
+#else
+  strncpy(tmp, in, n);	/* NetBSD delivers host byte order strings */  
+#endif
   tmp[n] = '\0';
   trim(out, tmp);
 }
 
-// Function for printing ASCII byte-swapped strings, skipping white
-// space. Please note that this is needed on both big- and
-// little-endian hardware.
-void printswap(char *output, char *in, unsigned int n){
-  formatdriveidstring(output, in, n);
+void infofound(char *output) {
   if (*output)
     pout("%s\n", output);
   else
     pout("[No Information Found]\n");
 }
 
+
 /* For the given Command Register (CR) and Features Register (FR), attempts
  * to construct a string that describes the contents of the Status
  * Register (ST) and Error Register (ER).  The string is dynamically allocated
@@ -439,7 +445,17 @@ uint64_t determine_capacity(struct ata_identify_device *drive, char *pstring){
   unsigned short lba_64         = drive->words088_255[103-88];
   uint64_t capacity_short=0, capacity=0, threedigits, power_of_ten;
   int started=0,k=1000000000;
-  
+  char separator=',';
+
+  // get correct character to use as thousands separator
+#ifdef HAVE_LOCALE_H
+  struct lconv *currentlocale=NULL;
+  setlocale (LC_ALL, "");
+  currentlocale=localeconv();
+  if (*(currentlocale->thousands_sep))
+    separator=*(currentlocale->thousands_sep);
+#endif // #ifdef HAVE_LOCALE_H
+
   // if drive supports LBA addressing, determine 32-bit LBA capacity
   if (capabilities_0 & 0x0200) {
     capacity_short = (unsigned int)sects_32 << 16 | 
@@ -460,8 +476,7 @@ uint64_t determine_capacity(struct ata_identify_device *drive, char *pstring){
   // turn sectors into bytes
   capacity_short = (capacity *= 512);
   
-  // print with comma separators.  I know this is anglo-centric:
-  // tell me what to change to use LOCALE if you want.
+  // print with locale-specific separators (default is comma)
   power_of_ten =  k;
   power_of_ten *= k;
   
@@ -470,7 +485,7 @@ uint64_t determine_capacity(struct ata_identify_device *drive, char *pstring){
     capacity   -= threedigits*power_of_ten;
     if (started)
       // we have already printed some digits
-      pstring += sprintf(pstring, ",%03"PRIu64, threedigits);
+      pstring += sprintf(pstring, "%c%03"PRIu64, separator, threedigits);
     else if (threedigits || k==6) {
       // these are the first digits that we are printing
       pstring += sprintf(pstring, "%"PRIu64, threedigits);
@@ -489,23 +504,30 @@ void ataPrintDriveInfo (struct ata_identify_device *drive){
   char unknown[64], timedatetz[DATEANDEPOCHLEN];
   unsigned short minorrev;
   char model[64], serial[64], firm[64], capacity[64];
-  
+
+  // format drive information (with byte swapping as needed)
+  formatdriveidstring(model, (char *)drive->model,40);
+  formatdriveidstring(serial, (char *)drive->serial_no,20);
+  formatdriveidstring(firm, (char *)drive->fw_rev,8);
 
   // print out model, serial # and firmware versions  (byte-swap ASCI strings)
+  drivetype=lookupdrive(model, firm);
+
+  // Print model family if known
+  if (drivetype>=0 && knowndrives[drivetype].modelfamily)
+    pout("Model Family:     %s\n", knowndrives[drivetype].modelfamily);
+
   pout("Device Model:     ");
-  printswap(model, (char *)drive->model,40);
-
+  infofound(model);
   pout("Serial Number:    ");
-  printswap(serial, (char *)drive->serial_no,20);
-
+  infofound(serial);
   pout("Firmware Version: ");
-  printswap(firm, (char *)drive->fw_rev,8);
+  infofound(firm);
 
   if (determine_capacity(drive, capacity))
     pout("User Capacity:    %s bytes\n", capacity);
   
   // See if drive is recognized
-  drivetype=lookupdrive(model, firm);
   pout("Device is:        %s\n", drivetype<0?
        "Not in smartctl database [for details use: -P showall]":
        "In smartctl database [for details use: -P show]");
@@ -1456,9 +1478,17 @@ int ataPrintMain (int fd){
       pout("                  Checking to be sure by trying SMART RETURN STATUS command.\n");
       isenabled=ataDoesSmartWork(fd);
     }
-    else
+    else {
       pout("SMART support is: Available - device has SMART capability.\n");
-    
+#ifdef HAVE_ATA_IDENTIFY_IS_CACHED
+      if (ata_identify_is_cached(fd)) {
+        pout("                  %sabled status cached by OS, trying SMART RETURN STATUS cmd.\n",
+                    (isenabled?"En":"Dis"));
+        isenabled=ataDoesSmartWork(fd);
+      }
+#endif
+    }
+
     if (isenabled)
       pout("SMART support is: Enabled\n");
     else {
