@@ -3,8 +3,8 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
- * Copyright (C) 2004 Christian Franke
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2004-6 Christian Franke
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,11 +20,7 @@
 #ifndef INT64_H_
 #define INT64_H_
 
-#define INT64_H_CVSID "$Id: int64.h,v 1.8 2004/09/07 22:10:44 shattered Exp $\n"
-
-#ifndef CONFIG_H_CVSID
-#include "config.h"
-#endif
+#define INT64_H_CVSID "$Id: int64.h,v 1.13 2006/04/12 14:54:28 ballen4705 Exp $\n"
 
 // 64 bit integer typedefs
 
@@ -44,6 +40,16 @@
 // for MSVC 6.0
 typedef          __int64    int64_t;
 typedef unsigned __int64   uint64_t;
+#else
+// for systems with above includes missing (like ix86-pc-linux-gnulibc1),
+// default to GCC if types are undefined in types.h
+#include <sys/types.h>
+#ifndef HAVE_INT64_T
+typedef          long long  int64_t;
+#endif
+#ifndef HAVE_UINT64_T
+typedef unsigned long long uint64_t;
+#endif
 #endif // _WIN32 && _MSC_VER
 #endif // HAVE_SYS_INT_TYPES_H
 #endif // HAVE_SYS_INTTYPES_H
@@ -75,9 +81,11 @@ typedef unsigned __int64   uint64_t;
 
 
 #if defined(_WIN32) && defined(_MSC_VER)
-// for MSVC 6.0: "unsigned __int64 -> double" conversion not implemented
-// replacement function implemented in os_win32/int64_vc6.c
-double uint64_to_double(unsigned __int64 ull);
+// for MSVC 6.0: "unsigned __int64 -> double" conversion not implemented (why?-)
+__inline double uint64_to_double(uint64_t ull) {
+  return ((int64_t)ull >= 0 ? (double)(int64_t)ull :
+    ((double)(int64_t)(ull - 9223372036854775808UI64)) + 9223372036854775808.0);
+}
 #else
 #define uint64_to_double(ull) ((double)(ull))
 #endif // _WIN32 && _MSC_VER
