@@ -4,7 +4,7 @@
  * Home page of code is: http://smartmontools.sourceforge.net
  * Address of support mailing list: smartmontools-support@lists.sourceforge.net
  *
- * Copyright (C) 2003-4 Philip Williams, Bruce Allen
+ * Copyright (C) 2003-6 Philip Williams, Bruce Allen
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,11 +20,14 @@
 #ifndef KNOWNDRIVES_H_
 #define KNOWNDRIVES_H_
 
-#define KNOWNDRIVES_H_CVSID "$Id: knowndrives.h,v 1.10 2004/01/02 16:05:25 ballen4705 Exp $\n"
+#define KNOWNDRIVES_H_CVSID "$Id: knowndrives.h,v 1.16 2006/04/05 19:50:07 chrfranke Exp $\n"
 
 /* Structure used to store settings for specific drives in knowndrives[]. The
  * elements are used in the following ways:
  *
+ *  modelfamily     Informal string about the model family/series of a 
+ *                  device. Set to NULL if no info (apart from device id)
+ *                  known.
  *  modelregexp     POSIX regular expression to match the model of a device.
  *                  This should never be NULL (except to terminate the
  *                  knowndrives array).
@@ -47,6 +50,7 @@
  *                  make the output more informative.
  */
 typedef struct drivesettings_s {
+  const char * const modelfamily;
   const char * const modelregexp;
   const char * const firmwareregexp;
   const char * const warningmsg;
@@ -66,7 +70,12 @@ int lookupdrive(const char *model, const char *firmware);
 void showpresets(const struct ata_identify_device *drive);
 
 // Shows all presets for drives in knowndrives[].
-void showallpresets(void);
+// Returns <0 on syntax error in regular expressions.
+int showallpresets(void);
+
+// Shows all matching presets for a drive in knowndrives[].
+// Returns # matching entries.
+int showmatchingpresets(const char *model, const char *firmware);
 
 // Sets preset vendor attribute options in opts by finding the entry
 // (if any) for the given drive in knowndrives[].  Values that have
