@@ -4,7 +4,7 @@
  * Home page of code is: http://smartmontools.sourceforge.net
  *
  * Copyright (C) YEAR YOUR_NAME <smartmontools-support@lists.sourceforge.net>
- * Copyright (C) 2003-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2003-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -46,27 +46,46 @@
      about checking your code into the smartmontools CVS archive.
 */
 
-// These are needed to define prototypes for the functions defined below
+/*
+ Developer's note: for testing this file, use an unsupported system,
+ for example: ./configure --build=rs6000-ibm-aix && make
+*/
+
+
+// This is needed for the various HAVE_* macros and PROJECT_* macros.
+#include "config.h"
+
+// These are needed to define prototypes and structures for the
+// functions defined below
+#include "int64.h"
 #include "atacmds.h"
 #include "scsicmds.h"
 #include "utility.h"
 
-// This is to include whatever prototypes you define in os_generic.h
+// This is to include whatever structures and prototypes you define in
+// os_generic.h
 #include "os_generic.h"
 
-// Needed by '-V' option (CVS versioning) of smartd/smartctl
-const char *os_XXXX_c_cvsid="$Id: os_generic.c,v 1.17 2004/08/18 21:27:18 ballen4705 Exp $" \
-ATACMDS_H_CVSID OS_XXXX_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
+// Needed by '-V' option (CVS versioning) of smartd/smartctl.  You
+// should have one *_H_CVSID macro appearing below for each file
+// appearing with #include "*.h" above.  Please list these (below) in
+// alphabetic/dictionary order.
+const char *os_XXXX_c_cvsid="$Id: os_generic.c,v 1.21 2006/04/12 14:54:28 ballen4705 Exp $" \
+ATACMDS_H_CVSID CONFIG_H_CVSID INT64_H_CVSID OS_GENERIC_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
 
 
-// Please eliminate the following block: both the two #includes and
+// This is here to prevent compiler warnings for unused arguments of
+// functions.
+#define ARGUSED(x) ((void)(x))
+
+// Please eliminate the following block: both the #include and
 // the 'unsupported()' function.  They are only here to warn
 // unsuspecting users that their Operating System is not supported! If
 // you wish, you can use a similar warning mechanism for any of the
 // functions in this file that you can not (or choose not to)
 // implement.
 
-#include "config.h"
+
 #ifdef HAVE_UNAME
 #include <sys/utsname.h>
 #endif
@@ -136,6 +155,7 @@ void print_smartctl_examples(){
 // tries to guess device type given the name (a path).  See utility.h
 // for return values.
 int guess_device_type (const char* dev_name) {
+  ARGUSED(dev_name);
   unsupported();
   return CONTROLLER_UNKNOWN;
 }
@@ -147,6 +167,8 @@ int guess_device_type (const char* dev_name) {
 // the case N==0, no arrays are allocated because the array of 0
 // pointers has zero length, equivalent to calling malloc(0).
 int make_device_names (char*** devlist, const char* name) {
+  ARGUSED(devlist);
+  ARGUSED(name);
   unsupported();
   return 0;
 }
@@ -158,6 +180,8 @@ int make_device_names (char*** devlist, const char* name) {
 // can not open the device (permission denied, does not exist, etc)
 // set errno as open() does and return <0.
 int deviceopen(const char *pathname, char *type){
+  ARGUSED(pathname);
+  ARGUSED(type);
   unsupported();
   return -1;
 }
@@ -165,6 +189,7 @@ int deviceopen(const char *pathname, char *type){
 // Like close().  Acts only on integer handles returned by
 // deviceopen() above.
 int deviceclose(int fd){
+  ARGUSED(fd);
   unsupported();
   return 0;
 }
@@ -185,20 +210,36 @@ int deviceclose(int fd){
 //   0 if the command succeeded and disk SMART status is "OK"
 //   1 if the command succeeded and disk SMART status is "FAILING"
 int ata_command_interface(int fd, smart_command_set command, int select, char *data){
+  ARGUSED(fd);
+  ARGUSED(command);
+  ARGUSED(select);
+  ARGUSED(data);
   unsupported();
   return -1;
 }
 
 int marvell_command_interface(int fd, smart_command_set command, int select, char *data){
+  ARGUSED(fd);
+  ARGUSED(command);
+  ARGUSED(select);
+  ARGUSED(data);
   unsupported();
   return -1;
 }
 
-// Interface to ATA devices behind 3ware escalade RAID controller
-// cards.  Same description as ata_command_interface() above except
-// that 0 <= disknum <= 15 specifies the ATA disk attached to the
-// controller.
-int escalade_command_interface(int fd, int disknum, int escalade_type, smart_command_set command, int select, char *data){
+// Interface to ATA devices behind 3ware escalade/apache RAID
+// controller cards.  Same description as ata_command_interface()
+// above except that 0 <= disknum <= 15 specifies the ATA disk
+// attached to the controller, and controller_type specifies the
+// precise type of 3ware controller.  See os_linux.c
+int escalade_command_interface(int fd, int disknum, int controller_type, smart_command_set command, int select, char *data){
+  ARGUSED(fd);
+  ARGUSED(disknum);
+  ARGUSED(controller_type);
+  ARGUSED(command);
+  ARGUSED(select);
+  ARGUSED(data);
+
   unsupported();
   return -1;
 }
@@ -206,6 +247,9 @@ int escalade_command_interface(int fd, int disknum, int escalade_type, smart_com
 #include <errno.h>
 // Interface to SCSI devices.  See os_linux.c
 int do_scsi_cmnd_io(int fd, struct scsi_cmnd_io * iop, int report) {
+  ARGUSED(fd);
+  ARGUSED(iop);
+  ARGUSED(report);
   unsupported();
   return -ENOSYS;
 }
