@@ -3,8 +3,8 @@
  * 
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2003-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
- * Copyright (C) 2003-4 Doug Gilbert <dougg@torque.net>
+ * Copyright (C) 2003-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2003-6 Doug Gilbert <dougg@torque.net>
  *
  *  Parts of this file are derived from code that was
  *
@@ -58,6 +58,8 @@
 #include <sys/sysmacros.h>
 #endif
 
+#include "config.h"
+#include "int64.h"
 #include "atacmds.h"
 #include "os_linux.h"
 #include "scsicmds.h"
@@ -70,10 +72,10 @@ typedef unsigned long long u8;
 
 #define ARGUSED(x) ((void)(x))
 
-static const char *filenameandversion="$Id: os_linux.c,v 1.72 2004/08/25 03:12:40 ballen4705 Exp $";
+static const char *filenameandversion="$Id: os_linux.c,v 1.82 2006/04/12 16:28:56 ballen4705 Exp $";
 
-const char *os_XXXX_c_cvsid="$Id: os_linux.c,v 1.72 2004/08/25 03:12:40 ballen4705 Exp $" \
-ATACMDS_H_CVSID OS_XXXX_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
+const char *os_XXXX_c_cvsid="$Id: os_linux.c,v 1.82 2006/04/12 16:28:56 ballen4705 Exp $" \
+ATACMDS_H_CVSID CONFIG_H_CVSID INT64_H_CVSID OS_LINUX_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
 
 // to hold onto exit code for atexit routine
 extern int exitstatus;
@@ -203,13 +205,15 @@ void print_smartctl_examples(){
   printf("=================================================== SMARTCTL EXAMPLES =====\n\n");
 #ifdef HAVE_GETOPT_LONG
   printf(
-         "  smartctl -a /dev/hda                       (Prints all SMART information)\n\n"
+         "  smartctl --all /dev/hda                    (Prints all SMART information)\n\n"
          "  smartctl --smart=on --offlineauto=on --saveauto=on /dev/hda\n"
          "                                              (Enables SMART on first disk)\n\n"
-         "  smartctl -t long /dev/hda              (Executes extended disk self-test)\n\n"
+         "  smartctl --test=long /dev/hda          (Executes extended disk self-test)\n\n"
          "  smartctl --attributes --log=selftest --quietmode=errorsonly /dev/hda\n"
          "                                      (Prints Self-Test & Attribute errors)\n"
-         "  smartctl -a --device=3ware,2 /dev/sda\n"
+         "  smartctl --all --device=3ware,2 /dev/sda\n"
+         "  smartctl --all --device=3ware,2 /dev/twe0\n"
+         "  smartctl --all --device=3ware,2 /dev/twa0\n"
          "          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
          );
 #else
@@ -220,6 +224,8 @@ void print_smartctl_examples(){
          "  smartctl -A -l selftest -q errorsonly /dev/hda\n"
          "                                      (Prints Self-Test & Attribute errors)\n"
          "  smartctl -a -d 3ware,2 /dev/sda\n"
+         "  smartctl -a -d 3ware,2 /dev/twa0\n"
+         "  smartctl -a -d 3ware,2 /dev/twe0\n"
          "          (Prints all SMART info for 3rd ATA disk on 3ware RAID controller)\n"
          );
 #endif
@@ -637,7 +643,8 @@ int ata_command_interface(int device, smart_command_set command, int select, cha
 #define SG_IO_PRESENT_YES 1
 #define SG_IO_PRESENT_NO 2
 
-static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report);
+static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report,
+			 int unknown);
 static int sisc_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report);
 
 static int sg_io_state = SG_IO_PRESENT_UNKNOWN;
@@ -646,7 +653,8 @@ static int sg_io_state = SG_IO_PRESENT_UNKNOWN;
  * function uses the SG_IO ioctl. Return 0 if command issued successfully
  * (various status values should still be checked). If the SCSI command
  * cannot be issued then a negative errno value is returned. */
-static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report)
+static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report,
+			 int unknown)
 {
 #ifndef SG_IO
     ARGUSED(dev_fd); ARGUSED(iop); ARGUSED(report);
@@ -708,7 +716,7 @@ static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report)
     iop->scsi_status = 0;
     iop->resid = 0;
     if (ioctl(dev_fd, SG_IO, &io_hdr) < 0) {
-        if (report)
+        if (report && (! unknown))
             pout("  SG_IO ioctl failed, errno=%d [%s]\n", errno,
 		 strerror(errno));
         return -errno;
@@ -746,10 +754,10 @@ static int sg_io_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report)
         if (0 != masked_driver_status) {
             if (LSCSI_DRIVER_TIMEOUT == masked_driver_status)
                 return -ETIMEDOUT;
-	    else
-		return -EIO;	/* catch all */
+	    else if (LSCSI_DRIVER_SENSE != masked_driver_status)
+		return -EIO;
 	}
-        if (LSCSI_DRIVER_SENSE == (io_hdr.driver_status & 0xf))
+        if (LSCSI_DRIVER_SENSE == masked_driver_status)
             iop->scsi_status = SCSI_STATUS_CHECK_CONDITION;
         iop->resp_sense_len = io_hdr.sb_len_wr;
         if ((SCSI_STATUS_CHECK_CONDITION == iop->scsi_status) && 
@@ -917,7 +925,7 @@ int do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report)
     switch (sg_io_state) {
     case SG_IO_PRESENT_UNKNOWN:
         /* ignore report argument */
-	if (0 == (res = sg_io_cmnd_io(dev_fd, iop, 0))) {
+	if (0 == (res = sg_io_cmnd_io(dev_fd, iop, report, 1))) {
 	    sg_io_state = SG_IO_PRESENT_YES;
 	    return 0;
 	} else if ((-ENODEV == res) || (-EACCES == res) || (-EPERM == res))
@@ -927,7 +935,7 @@ int do_scsi_cmnd_io(int dev_fd, struct scsi_cmnd_io * iop, int report)
     case SG_IO_PRESENT_NO:
 	return sisc_cmnd_io(dev_fd, iop, report);
     case SG_IO_PRESENT_YES:
-	return sg_io_cmnd_io(dev_fd, iop, report);
+	return sg_io_cmnd_io(dev_fd, iop, report, 0);
     default:
 	pout(">>>> do_scsi_cmnd_io: bad sg_io_state=%d\n", sg_io_state); 
 	sg_io_state = SG_IO_PRESENT_UNKNOWN;
@@ -1117,6 +1125,7 @@ int escalade_command_interface(int fd, int disknum, int escalade_type, smart_com
   case PIDENTIFY:
     // 3WARE controller can NOT have packet device internally
     pout("WARNING - NO DEVICE FOUND ON 3WARE CONTROLLER (disk %d)\n", disknum);
+    pout("Note: /dev/sdX many need to be replaced with /dev/tweN or /dev/twaN\n");
     errno=ENODEV;
     return -1;
   case ENABLE:
@@ -1272,7 +1281,7 @@ int marvell_command_interface(int device,
   
   int copydata = 0;
   mvsata_scsi_cmd  smart_command;
-  unsigned char *buff = &smart_command.cmd[6];
+  unsigned char *buff = (unsigned char *)&smart_command.cmd[6];
   // See struct hd_drive_cmd_hdr in hdreg.h
   // buff[0]: ATA COMMAND CODE REGISTER
   // buff[1]: ATA SECTOR NUMBER REGISTER
@@ -1356,8 +1365,12 @@ int marvell_command_interface(int device,
     return 0;
   }
 
+  // Always succeed on a SMART status, as a disk that failed returned  
+  // buff[4]=0xF4, buff[5]=0x2C, i.e. "Bad SMART status" (see below).
+  if (command == STATUS)
+    return 0;
   //Data returned is starting from 0 offset  
-  if (command == STATUS || command == STATUS_CHECK)
+  if (command == STATUS_CHECK)
   {
     // Cyl low and Cyl high unchanged means "Good SMART status"
     if (buff[4] == 0x4F && buff[5] == 0xC2)
