@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2004 Christian Franke <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2004-6 Christian Franke <smartmontools-support@lists.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,10 +17,10 @@
  */
 
 #include "config.h"
+#include "int64.h"
 #include "atacmds.h"
 #include "extern.h"
-extern smartmonctrl * con; // con->permissive
-#include "int64.h"
+extern smartmonctrl * con; // con->permissive,reportataioctl
 #include "scsicmds.h"
 #include "utility.h"
 extern int64_t bytes; // malloc() byte count
@@ -39,17 +39,82 @@ extern int64_t bytes; // malloc() byte count
 #define ARGUSED(x) ((void)(x))
 
 // Needed by '-V' option (CVS versioning) of smartd/smartctl
-const char *os_XXXX_c_cvsid="$Id: os_win32.c,v 1.17 2004/08/18 19:27:44 likewise Exp $"
+const char *os_XXXX_c_cvsid="$Id: os_win32.c,v 1.33 2006/04/07 15:02:19 chrfranke Exp $"
 ATACMDS_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
 
 
+#ifndef HAVE_GET_OS_VERSION_STR
+#error define of HAVE_GET_OS_VERSION_STR missing in config.h
+#endif
+
+// Return build host and OS version as static string
+const char * get_os_version_str()
+{
+	static char vstr[sizeof(SMARTMONTOOLS_BUILD_HOST)-3-1+sizeof("-2003r2-sp2.1")+13];
+	char * const vptr = vstr+sizeof(SMARTMONTOOLS_BUILD_HOST)-3-1;
+	const int vlen = sizeof(vstr)-(sizeof(SMARTMONTOOLS_BUILD_HOST)-3);
+
+	OSVERSIONINFOEXA vi;
+	const char * w;
+
+	// remove "-pc" to avoid long lines
+	assert(!strncmp(SMARTMONTOOLS_BUILD_HOST+5, "pc-", 3));
+	strcpy(vstr, "i686-"); strcpy(vstr+5, SMARTMONTOOLS_BUILD_HOST+5+3);
+	assert(vptr == vstr+strlen(vstr) && vptr+vlen+1 == vstr+sizeof(vstr));
+
+	memset(&vi, 0, sizeof(vi));
+	vi.dwOSVersionInfoSize = sizeof(vi);
+	if (!GetVersionExA((OSVERSIONINFOA *)&vi)) {
+		memset(&vi, 0, sizeof(vi));
+		vi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOA);
+		if (!GetVersionExA((OSVERSIONINFOA *)&vi))
+			return vstr;
+	}
+
+	if (vi.dwPlatformId > 0xff || vi.dwMajorVersion > 0xff || vi.dwMinorVersion > 0xff)
+		return vstr;
+
+	switch (vi.dwPlatformId << 16 | vi.dwMajorVersion << 8 | vi.dwMinorVersion) {
+	  case VER_PLATFORM_WIN32_WINDOWS<<16|0x0400| 0:
+		w = (vi.szCSDVersion[1] == 'B' ||
+		     vi.szCSDVersion[1] == 'C'     ? "95-osr2" : "95");    break;
+	  case VER_PLATFORM_WIN32_WINDOWS<<16|0x0400|10:
+		w = (vi.szCSDVersion[1] == 'A'     ? "98se"    : "98");    break;
+	  case VER_PLATFORM_WIN32_WINDOWS<<16|0x0400|90: w = "me";     break;
+	//case VER_PLATFORM_WIN32_NT     <<16|0x0300|51: w = "nt3.51"; break;
+	  case VER_PLATFORM_WIN32_NT     <<16|0x0400| 0: w = "nt4";    break;
+	  case VER_PLATFORM_WIN32_NT     <<16|0x0500| 0: w = "2000";   break;
+	  case VER_PLATFORM_WIN32_NT     <<16|0x0500| 1:
+		w = (!GetSystemMetrics(87/*SM_MEDIACENTER*/) ?   "xp"
+		                                             :   "xp-mc"); break;
+	  case VER_PLATFORM_WIN32_NT     <<16|0x0500| 2:
+		w = (!GetSystemMetrics(89/*SM_SERVERR2*/) ?      "2003"
+		                                             :   "2003r2"); break;
+	  case VER_PLATFORM_WIN32_NT     <<16|0x0600| 0: w = "vista";  break;
+	  default: w = 0; break;
+	}
+
+	if (!w)
+		snprintf(vptr, vlen, "-%s%lu.%lu",
+			(vi.dwPlatformId==VER_PLATFORM_WIN32_NT ? "nt" : "9x"),
+			vi.dwMajorVersion, vi.dwMinorVersion);
+	else if (vi.wServicePackMinor)
+		snprintf(vptr, vlen, "-%s-sp%u.%u", w, vi.wServicePackMajor, vi.wServicePackMinor);
+	else if (vi.wServicePackMajor)
+		snprintf(vptr, vlen, "-%s-sp%u", w, vi.wServicePackMajor);
+	else
+		snprintf(vptr, vlen, "-%s", w);
+	return vstr;
+}
+
+
 static int ata_open(int drive);
 static void ata_close(int fd);
-static unsigned ata_scan(void);
+static int ata_scan(unsigned long * drives);
 
 static int aspi_open(unsigned adapter, unsigned id);
 static void aspi_close(int fd);
-static unsigned long aspi_scan(void);
+static int aspi_scan(unsigned long * drives);
 
 
 static int is_permissive()
@@ -87,40 +152,30 @@ int guess_device_type (const char * dev_name)
 // others each contain null-terminated character strings.
 int make_device_names (char*** devlist, const char* type)
 {
-	unsigned long drives;
-	int i, j, n, sz, scsi;
+	unsigned long drives[3];
+	int i, j, n, nmax, sz;
 	const char * path;
 
+	drives[0] = drives[1] = drives[2] = 0;
 	if (!strcmp(type, "ATA")) {
 		// bit i set => drive i present
-		drives = ata_scan();
+		n = ata_scan(drives);
 		path = "/dev/hda";
-		scsi = 0;
+		nmax = 10;
 	}
 	else if (!strcmp(type, "SCSI")) {
 		// bit i set => drive with ID (i & 0x7) on adapter (i >> 3) present
-		drives = aspi_scan();
+		n = aspi_scan(drives);
 		path = "/dev/scsi00";
-		scsi = 1;
+		nmax = 10*8;
 	}
 	else
 		return -1;
 
-	if (!drives)
-		return 0;
-
-	// Count #drives
-	n = 0;
-	for (i = 0; i < 32; i++) {
-		if (drives & (1 << i))
-			n++;
-	}
-	assert(n > 0);
-	if (n == 0)
+	if (n <= 0)
 		return 0;
 
 	// Alloc devlist
-	assert(scsi || n <= 9);
 	sz = n * sizeof(char **);
 	*devlist = (char **)malloc(sz); bytes += sz;
 
@@ -130,15 +185,16 @@ int make_device_names (char*** devlist, const char* type)
 		sz = strlen(path)+1;
 		s = (char *)malloc(sz); bytes += sz;
 		strcpy(s, path);
-		while (j < 32 && !(drives & (1 << j)))
+		while (j < nmax && !(drives[j >> 5] & (1L << (j & 0x1f))))
 			j++;
-		assert(j < 32);
-		if (!scsi) {
+		assert(j < nmax);
+		if (nmax <= 10) {
 			assert(j <= 9);
 			s[sz-2] += j; // /dev/hd[a-j]
 		}
 		else {
-			s[sz-3] += (j >> 3);  // /dev/scsi[0-3].
+			assert((j >> 3) <= 9);
+			s[sz-3] += (j >> 3);  // /dev/scsi[0-9].....
 			s[sz-2] += (j & 0x7); //          .....[0-7]
 		}
 		(*devlist)[i] = s;
@@ -317,6 +373,13 @@ static void print_ide_regs(const IDEREGS * r, int out)
 	r->bSectorCountReg, r->bSectorNumberReg, r->bCylLowReg, r->bCylHighReg, r->bDriveHeadReg);
 }
 
+static void print_ide_regs_io(const IDEREGS * ri, const IDEREGS * ro)
+{
+	pout("    Input : "); print_ide_regs(ri, 0);
+	if (ro) {
+		pout("    Output: "); print_ide_regs(ro, 1);
+	}
+}
 
 /////////////////////////////////////////////////////////////////////////////
 
@@ -329,6 +392,7 @@ static int smart_ioctl(HANDLE hdevice, int drive, IDEREGS * regs, char * data, u
 	const SENDCMDOUTPARAMS * outpar;
 	DWORD code, num_out;
 	unsigned int size_out;
+	const char * name;
 
 	assert(SMART_SEND_DRIVE_COMMAND == 0x07c084);
 	assert(SMART_RCV_DRIVE_DATA == 0x07c088);
@@ -343,34 +407,28 @@ static int smart_ioctl(HANDLE hdevice, int drive, IDEREGS * regs, char * data, u
 
 	assert(datasize == 0 || datasize == 512);
 	if (datasize) {
+		code = SMART_RCV_DRIVE_DATA; name = "SMART_RCV_DRIVE_DATA";
 		inpar.cBufferSize = size_out = 512;
-		code = SMART_RCV_DRIVE_DATA;
-	}
-	else if (regs->bFeaturesReg == ATA_SMART_STATUS) {
-		size_out = sizeof(IDEREGS); // ioctl returns new IDEREGS as data
-		code = SMART_SEND_DRIVE_COMMAND;
 	}
 	else {
-		size_out = 0;
-		code = SMART_SEND_DRIVE_COMMAND;
+		code = SMART_SEND_DRIVE_COMMAND; name = "SMART_SEND_DRIVE_COMMAND";
+		if (regs->bFeaturesReg == ATA_SMART_STATUS)
+			size_out = sizeof(IDEREGS); // ioctl returns new IDEREGS as data
+			// Note: cBufferSize must be 0 on Win9x
+		else
+			size_out = 0;
 	}
 
 	memset(&outbuf, 0, sizeof(outbuf));
 
-#ifdef _DEBUG
-	pout("DeviceIoControl(.,0x%lx,.,%lu,.,%lu,.,NULL)\n",
-		code, sizeof(SENDCMDINPARAMS)-1, sizeof(SENDCMDOUTPARAMS)-1 + size_out);
-	print_ide_regs(&inpar.irDriveRegs, 0);
-#endif
-	if (!DeviceIoControl(hdevice, code,
-             		&inpar, sizeof(SENDCMDINPARAMS)-1,
-               		outbuf, sizeof(SENDCMDOUTPARAMS)-1 + size_out,
-               		&num_out, NULL)) {
+	if (!DeviceIoControl(hdevice, code, &inpar, sizeof(SENDCMDINPARAMS)-1,
+		outbuf, sizeof(SENDCMDOUTPARAMS)-1 + size_out, &num_out, NULL)) {
 		// CAUTION: DO NOT change "regs" Parameter in this case, see ata_command_interface()
 		long err = GetLastError();
-#ifdef _DEBUG
-		pout("DeviceIoControl failed, Error=%ld\n", err);
-#endif
+		if (con->reportataioctl && (err != ERROR_INVALID_PARAMETER || con->reportataioctl > 1)) {
+			pout("  %s failed, Error=%ld\n", name, err);
+			print_ide_regs_io(regs, NULL);
+		}
 		errno = (   err == ERROR_INVALID_FUNCTION /*9x*/
 		         || err == ERROR_INVALID_PARAMETER/*NT/2K/XP*/ ? ENOSYS : EIO);
 		return -1;
@@ -378,25 +436,28 @@ static int smart_ioctl(HANDLE hdevice, int drive, IDEREGS * regs, char * data, u
 	// NOTE: On Win9x, inpar.irDriveRegs now contains the returned regs
 
 	outpar = (const SENDCMDOUTPARAMS *)outbuf;
-#ifdef _DEBUG
-	pout("DeviceIoControl returns %lu (%lu) bytes\n", num_out, outpar->cBufferSize);
-#endif
 
 	if (outpar->DriverStatus.bDriverError) {
-		pout("Error SMART IOCTL DriverError=0x%02x, IDEError=0x%02x\n",
-			outpar->DriverStatus.bDriverError, outpar->DriverStatus.bIDEError);
+		if (con->reportataioctl) {
+			pout("  %s failed, DriverError=0x%02x, IDEError=0x%02x\n", name,
+				outpar->DriverStatus.bDriverError, outpar->DriverStatus.bIDEError);
+			print_ide_regs_io(regs, NULL);
+		}
 		errno = EIO;
 		return -1;
 	}
 
+	if (con->reportataioctl > 1) {
+		pout("  %s suceeded, bytes returned: %lu (buffer %lu)\n", name,
+			num_out, outpar->cBufferSize);
+		print_ide_regs_io(regs, (regs->bFeaturesReg == ATA_SMART_STATUS ?
+			(const IDEREGS *)(outpar->bBuffer) : NULL));
+	}
+
 	if (datasize)
 		memcpy(data, outpar->bBuffer, 512);
-	else if (regs->bFeaturesReg == ATA_SMART_STATUS) {
+	else if (regs->bFeaturesReg == ATA_SMART_STATUS)
 		*regs = *(const IDEREGS *)(outpar->bBuffer);
-#ifdef _DEBUG
-		print_ide_regs(regs, 1);
-#endif
-	}
 
 	return 0;
 }
@@ -434,37 +495,185 @@ static int ide_pass_through_ioctl(HANDLE hdevice, IDEREGS * regs, char * data, u
 	unsigned int size = sizeof(ATA_PASS_THROUGH)-1 + datasize;
 	ATA_PASS_THROUGH * buf = (ATA_PASS_THROUGH *)VirtualAlloc(NULL, size, MEM_COMMIT, PAGE_READWRITE);
 	DWORD num_out;
+	const unsigned char magic = 0xcf;
 	assert(sizeof(ATA_PASS_THROUGH)-1 == 12);
 	assert(IOCTL_IDE_PASS_THROUGH == 0x04d028);
 
+	if (!buf) {
+		errno = ENOMEM;
+		return -1;
+	}
+
 	buf->IdeReg = *regs;
 	buf->DataBufferSize = datasize;
+	if (datasize)
+		buf->DataBuffer[0] = magic;
 
-#ifdef _DEBUG
-	pout("DeviceIoControl(.,0x%x,.,%u,.,%u,.,NULL)\n",
-		IOCTL_IDE_PASS_THROUGH, size, size);
-	print_ide_regs(&buf->IdeReg, 0);
-#endif
-
-	if (!DeviceIoControl(hdevice, IOCTL_IDE_PASS_THROUGH, 
+	if (!DeviceIoControl(hdevice, IOCTL_IDE_PASS_THROUGH,
 		buf, size, buf, size, &num_out, NULL)) {
 		long err = GetLastError();
-#ifdef _DEBUG
-		pout("DeviceIoControl failed, Error=%ld\n", err);
-#endif
-		VirtualFree(buf, size, MEM_RELEASE);
+		if (con->reportataioctl)
+			pout("  IOCTL_IDE_PASS_THROUGH failed, Error=%ld\n", err);
+		VirtualFree(buf, 0, MEM_RELEASE);
 		errno = (err == ERROR_INVALID_FUNCTION ? ENOSYS : EIO);
 		return -1;
 	}
 
-#ifdef _DEBUG
-	pout("DeviceIoControl returns %lu (%lu) bytes\n", num_out, buf->DataBufferSize);
-	print_ide_regs(&buf->IdeReg, 1);
-#endif
+	// Check ATA status
+	if (buf->IdeReg.bCommandReg/*Status*/ & 0x01) {
+		if (con->reportataioctl) {
+			pout("  IOCTL_IDE_PASS_THROUGH command failed:\n");
+			print_ide_regs_io(regs, &buf->IdeReg);
+		}
+		VirtualFree(buf, 0, MEM_RELEASE);
+		errno = EIO;
+		return -1;
+	}
 
-	if (datasize)
+	// Check and copy data
+	if (datasize) {
+		if (   num_out != size
+		    || (buf->DataBuffer[0] == magic && !nonempty(buf->DataBuffer+1, datasize-1))) {
+			if (con->reportataioctl) {
+				pout("  IOCTL_IDE_PASS_THROUGH output data missing (%lu, %lu)\n",
+					num_out, buf->DataBufferSize);
+				print_ide_regs_io(regs, &buf->IdeReg);
+			}
+			VirtualFree(buf, 0, MEM_RELEASE);
+			errno = EIO;
+			return -1;
+		}
 		memcpy(data, buf->DataBuffer, datasize);
-	VirtualFree(buf, size, MEM_RELEASE);
+	}
+
+	if (con->reportataioctl > 1) {
+		pout("  IOCTL_IDE_PASS_THROUGH suceeded, bytes returned: %lu (buffer %lu)\n",
+			num_out, buf->DataBufferSize);
+		print_ide_regs_io(regs, &buf->IdeReg);
+	}
+	*regs = buf->IdeReg;
+
+	// Caution: VirtualFree() fails if parameter "dwSize" is nonzero
+	VirtualFree(buf, 0, MEM_RELEASE);
+	return 0;
+}
+
+
+
+/////////////////////////////////////////////////////////////////////////////
+
+// ATA PASS THROUGH via SCSI PASS THROUGH for NT4 only
+// Only used for SMART commands not supported by SMART_* IOCTLs
+
+// Declarations from:
+// http://cvs.sourceforge.net/viewcvs.py/mingw/w32api/include/ddk/ntddscsi.h?rev=1.2
+
+#define IOCTL_SCSI_PASS_THROUGH \
+	CTL_CODE(IOCTL_SCSI_BASE, 0x0401, METHOD_BUFFERED, FILE_READ_ACCESS | FILE_WRITE_ACCESS)
+
+#define SCSI_IOCTL_DATA_OUT          0
+#define SCSI_IOCTL_DATA_IN           1
+#define SCSI_IOCTL_DATA_UNSPECIFIED  2
+// undocumented SCSI opcode to for ATA passthrough
+#define SCSIOP_ATA_PASSTHROUGH    0xCC
+
+typedef struct _SCSI_PASS_THROUGH {
+	USHORT  Length;
+	UCHAR  ScsiStatus;
+	UCHAR  PathId;
+	UCHAR  TargetId;
+	UCHAR  Lun;
+	UCHAR  CdbLength;
+	UCHAR  SenseInfoLength;
+	UCHAR  DataIn;
+	ULONG  DataTransferLength;
+	ULONG  TimeOutValue;
+	ULONG/*_PTR*/ DataBufferOffset;
+	ULONG  SenseInfoOffset;
+	UCHAR  Cdb[16];
+} SCSI_PASS_THROUGH, *PSCSI_PASS_THROUGH;
+
+
+/////////////////////////////////////////////////////////////////////////////
+
+static int ata_via_scsi_pass_through_ioctl(HANDLE hdevice, IDEREGS * regs, char * data, unsigned datasize)
+{
+	typedef struct {
+		SCSI_PASS_THROUGH spt;
+		ULONG Filler;
+		UCHAR ucSenseBuf[32];
+		UCHAR ucDataBuf[512];
+	} SCSI_PASS_THROUGH_WITH_BUFFERS;
+
+	SCSI_PASS_THROUGH_WITH_BUFFERS sb;
+	IDEREGS * cdbregs;
+	unsigned int size;
+	DWORD num_out;
+	const unsigned char magic = 0xcf;
+
+	assert(sizeof(SCSI_PASS_THROUGH) == 44);
+	assert(IOCTL_SCSI_PASS_THROUGH == 0x04d004);
+
+	memset(&sb, 0, sizeof(sb));
+	sb.spt.Length = sizeof(SCSI_PASS_THROUGH);
+	//sb.spt.PathId = 0;
+	sb.spt.TargetId = 1;
+	//sb.spt.Lun = 0;
+	sb.spt.CdbLength = 10; sb.spt.SenseInfoLength = 24;
+	sb.spt.TimeOutValue = 10;
+	sb.spt.SenseInfoOffset = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucSenseBuf);
+	size = offsetof(SCSI_PASS_THROUGH_WITH_BUFFERS, ucDataBuf);
+	sb.spt.DataBufferOffset = size;
+ 
+	if (datasize) {
+		if (datasize > sizeof(sb.ucDataBuf)) {
+			errno = EINVAL;
+			return -1;
+		}
+		sb.spt.DataIn = SCSI_IOCTL_DATA_IN;
+		sb.spt.DataTransferLength = datasize;
+		size += datasize;
+		sb.ucDataBuf[0] = magic;
+	}
+	else {
+		sb.spt.DataIn = SCSI_IOCTL_DATA_UNSPECIFIED;
+		//sb.spt.DataTransferLength = 0;
+	}
+
+	// Use pseudo SCSI command followed by registers
+	sb.spt.Cdb[0] = SCSIOP_ATA_PASSTHROUGH;
+	cdbregs = (IDEREGS *)(sb.spt.Cdb+2);
+	*cdbregs = *regs;
+
+	if (!DeviceIoControl(hdevice, IOCTL_SCSI_PASS_THROUGH,
+		&sb, size, &sb, size, &num_out, NULL)) {
+		long err = GetLastError();
+		if (con->reportataioctl)
+			pout("  ATA via IOCTL_SCSI_PASS_THROUGH failed, Error=%ld\n", err);
+		errno = (err == ERROR_INVALID_FUNCTION ? ENOSYS : EIO);
+		return -1;
+	}
+
+	// Cannot check ATA status, because command does not return IDEREGS
+
+	// Check and copy data
+	if (datasize) {
+		if (   num_out != size
+		    || (sb.ucDataBuf[0] == magic && !nonempty(sb.ucDataBuf+1, datasize-1))) {
+			if (con->reportataioctl) {
+				pout("  ATA via IOCTL_SCSI_PASS_THROUGH output data missing (%lu)\n", num_out);
+				print_ide_regs_io(regs, NULL);
+			}
+			errno = EIO;
+			return -1;
+		}
+		memcpy(data, sb.ucDataBuf, datasize);
+	}
+
+	if (con->reportataioctl > 1) {
+		pout("  ATA via IOCTL_SCSI_PASS_THROUGH suceeded, bytes returned: %lu\n", num_out);
+		print_ide_regs_io(regs, NULL);
+	}
 	return 0;
 }
 
@@ -472,7 +681,6 @@ static int ide_pass_through_ioctl(HANDLE hdevice, IDEREGS * regs, char * data, u
 /////////////////////////////////////////////////////////////////////////////
 
 static HANDLE h_ata_ioctl = 0;
-static int ide_pass_through_broken = 0;
 
 
 // Print SMARTVSD error message, return errno
@@ -483,27 +691,41 @@ static int smartvsd_error()
 	unsigned len;
 	if (!(5 <= (len = GetSystemDirectoryA(path, MAX_PATH)) && len < MAX_PATH/2))
 		return ENOENT;
+	// SMARTVSD.VXD present?
 	strcpy(path+len, "\\IOSUBSYS\\SMARTVSD.VXD");
 	if (!access(path, 0)) {
-#ifdef _DEBUG
-		pout("Driver \"%s\" not loaded,\n"
-		     " possibly no IDE/ATA devices present.\n", path);
-#endif
-		return ENOENT;
+		// Yes, standard IDE driver used?
+		HANDLE h;
+		if (   (h = CreateFileA("\\\\.\\ESDI_506",
+		             GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
+		             NULL, OPEN_EXISTING, 0, 0)) == INVALID_HANDLE_VALUE
+		    && GetLastError() == ERROR_FILE_NOT_FOUND                             ) {
+			pout("Standard IDE driver ESDI_506.PDR not used, or no IDE/ATA drives present.\n");
+			return ENOENT;
+		}
+		else {
+			if (h != INVALID_HANDLE_VALUE) // should not happen
+				CloseHandle(h);
+			pout("SMART driver SMARTVSD.VXD is installed, but not loaded.\n");
+			return ENOSYS;
+		}
 	}
-	// Some Windows versions install SMARTVSD.VXD in SYSTEM directory
-	// http://support.microsoft.com/default.aspx?scid=kb;en-us;265854
-	strcpy(path+len, "\\SMARTVSD.VXD");
-	if (!access(path, 0)) {
-		path[len] = 0;
-		pout("SMART driver is not properly installed,\n"
-		     " move SMARTVSD.VXD from \"%s\" to \"%s\\IOSUBSYS\"\n"
-		     " and reboot Windows.\n", path, path);
+	else {
+		// Some Windows versions install SMARTVSD.VXD in SYSTEM directory
+		// http://support.microsoft.com/default.aspx?scid=kb;en-us;265854
+		strcpy(path+len, "\\SMARTVSD.VXD");
+		if (!access(path, 0)) {
+			path[len] = 0;
+			pout("SMART driver is not properly installed,\n"
+				 " move SMARTVSD.VXD from \"%s\" to \"%s\\IOSUBSYS\"\n"
+				 " and reboot Windows.\n", path, path);
+		}
+		else {
+			path[len] = 0;
+			pout("SMARTVSD.VXD is missing in folder \"%s\\IOSUBSYS\".\n", path);
+		}
 		return ENOSYS;
 	}
-	path[len] = 0;
-	pout("SMARTVSD.VXD is missing in folder \"%s\\IOSUBSYS\".\n", path);
-	return ENOENT;
 }
 
 
@@ -541,10 +763,15 @@ static int ata_open(int drive)
 		NULL, OPEN_EXISTING, 0, 0)) == INVALID_HANDLE_VALUE) {
 		long err = GetLastError();	
 		pout("Cannot open device %s, Error=%ld\n", devpath, err);
-		if (win9x && err == ERROR_FILE_NOT_FOUND)
-			errno = smartvsd_error();
+		if (err == ERROR_FILE_NOT_FOUND)
+			errno = (win9x ? smartvsd_error() : ENOENT);
+		else if (err == ERROR_ACCESS_DENIED) {
+			if (!win9x)
+				pout("Administrator rights are necessary to access physical drives.\n");
+			errno = EACCES;
+		}
 		else
-			errno = (err == ERROR_FILE_NOT_FOUND ? ENOENT : EPERM);
+			errno = EIO;
 		h_ata_ioctl = 0;
 		return -1;
 	}
@@ -563,10 +790,11 @@ static int ata_open(int drive)
 		}
 	}
 
-#ifdef _DEBUG
-	pout("SMART_GET_VERSION (%ld bytes): Vers = %d.%d, Caps = 0x%lx, DeviceMap = 0x%02x\n",
-		num_out, vers.bVersion, vers.bRevision, vers.fCapabilities, vers.bIDEDeviceMap);
-#endif
+	if (con->reportataioctl > 1)
+		pout("%s: SMART_GET_VERSION (%ld bytes):\n"
+		     " Vers = %d.%d, Caps = 0x%lx, DeviceMap = 0x%02x\n",
+			devpath, num_out, vers.bVersion, vers.bRevision,
+			vers.fCapabilities, vers.bIDEDeviceMap);
 
 	// TODO: Check vers.fCapabilities here?
 
@@ -603,13 +831,12 @@ static void ata_close(int fd)
 }
 
 
-// Scan for ATA drives, return bitmask of drives present
+// Scan for ATA drives, fill bitmask of drives present, return #drives
 
-static unsigned ata_scan()
+static int ata_scan(unsigned long * drives)
 {
-	unsigned drives = 0;
 	int win9x = ((GetVersion() & 0x80000000) != 0);
-	int i;
+	int cnt = 0, i;
 
 	for (i = 0; i <= 9; i++) {
 		char devpath[30];
@@ -625,6 +852,8 @@ static unsigned ata_scan()
 		if ((h = CreateFileA(devpath,
 			GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,
 			NULL, OPEN_EXISTING, 0, 0)) == INVALID_HANDLE_VALUE) {
+			if (con->reportataioctl > 1)
+				pout(" %s: Open failed, Error=%ld\n", devpath, GetLastError());
 			if (win9x)
 				break; // SMARTVSD.VXD missing or no ATA devices
 			continue; // Disk not found or access denied (break;?)
@@ -634,6 +863,8 @@ static unsigned ata_scan()
 		memset(&vers, 0, sizeof(vers));
 		if (!DeviceIoControl(h, SMART_GET_VERSION,
 			NULL, 0, &vers, sizeof(vers), &num_out, NULL)) {
+			if (con->reportataioctl)
+				pout(" %s: SMART_GET_VERSION failed, Error=%ld\n", devpath, GetLastError());
 			CloseHandle(h);
 			if (win9x)
 				break; // Should not happen
@@ -641,48 +872,60 @@ static unsigned ata_scan()
 		}
 		CloseHandle(h);
 
+		if (con->reportataioctl)
+			pout(" %s: SMART_GET_VERSION (%ld bytes):\n"
+			     "  Vers = %d.%d, Caps = 0x%lx, DeviceMap = 0x%02x\n",
+				devpath, num_out, vers.bVersion, vers.bRevision,
+				vers.fCapabilities, vers.bIDEDeviceMap);
+
 		if (win9x) {
 			// Check ATA device presence, remove ATAPI devices
-			drives = (vers.bIDEDeviceMap & 0xf) & ~((vers.bIDEDeviceMap >> 4) & 0xf);
+			drives[0] = (vers.bIDEDeviceMap & 0xf) & ~((vers.bIDEDeviceMap >> 4) & 0xf);
+			cnt = (drives[0]&1) + ((drives[0]>>1)&1) + ((drives[0]>>2)&1) + ((drives[0]>>3)&1);
 			break;
 		}
 
 		// ATA drive exists and driver supports SMART ioctl
-		drives |= 1 << i;
+		drives[0] |= (1L << i);
+		cnt++;
 	}
 
-	return drives;
+	return cnt;
 }
 
 
 /////////////////////////////////////////////////////////////////////////////
 
 // Interface to ATA devices.  See os_linux.c
-int marvell_command_interface(int fd, smart_command_set command, int select, char * data)
-{ return -1; }
-
 int ata_command_interface(int fd, smart_command_set command, int select, char * data)
 {
 	IDEREGS regs;
-	int copydata;
+	int copydata, try_ioctl;
 
 	if (!(0 <= fd && fd <= 3)) {
-	  errno = EBADF;
-	  return -1;
+		errno = EBADF;
+		return -1;
 	}
 
-	// CMD,CYL default to SMART, changed by P?IDENTIFY
+	// CMD,CYL default to SMART, changed by P?IDENTIFY and CHECK_POWER_MODE
 	memset(&regs, 0, sizeof(regs));
 	regs.bCommandReg = ATA_SMART_CMD;
 	regs.bCylHighReg = SMART_CYL_HI; regs.bCylLowReg = SMART_CYL_LOW;
 	copydata = 0;
+	try_ioctl = 0x01; // 0x01=SMART_*, 0x02=IDE_PASS_THROUGH [, 0x04=ATA_PASS_THROUGH]
 
 	switch (command) {
-	  case CHECK_POWER_MODE:
 	  case WRITE_LOG:
-		// TODO. Not supported by SMART IOCTL
+		// TODO. Not supported by SMART IOCTL (no data out ioctl available),
+		//  also not supported by IOCTL_IDE_PASS_THROUGH (data out not working)
 		errno = ENOSYS;
 		return -1;
+	  case CHECK_POWER_MODE:
+		regs.bCommandReg = ATA_CHECK_POWER_MODE;
+		regs.bCylLowReg = regs.bCylHighReg = 0;
+		try_ioctl = 0x02; // IOCTL_IDE_PASS_THROUGH
+		// Note: returns SectorCountReg in data[0]
+		break;
 	  case READ_VALUES:
 		regs.bFeaturesReg = ATA_SMART_READ_VALUES;
 		regs.bSectorNumberReg = regs.bSectorCountReg = 1;
@@ -697,9 +940,13 @@ int ata_command_interface(int fd, smart_command_set command, int select, char *
 		regs.bFeaturesReg = ATA_SMART_READ_LOG_SECTOR;
 		regs.bSectorNumberReg = select;
 		regs.bSectorCountReg = 1;
+		// Read log only supported on Win9x, retry with pass through command
+		try_ioctl = 0x03; // SMART_RCV_DRIVE_DATA, then IOCTL_IDE_PASS_THROUGH
 		copydata = 1;
 		break;
 	  case IDENTIFY:
+		// Note: WinNT4/2000/XP return identify data cached during boot
+		// (true for SMART_RCV_DRIVE_DATA and IOCTL_IDE_PASS_THROUGH)
 		regs.bCommandReg = ATA_IDENTIFY_DEVICE;
 		regs.bCylLowReg = regs.bCylHighReg = 0;
 		regs.bSectorCountReg = 1;
@@ -734,6 +981,8 @@ int ata_command_interface(int fd, smart_command_set command, int select, char *
 	  case IMMEDIATE_OFFLINE:
 		regs.bFeaturesReg = ATA_SMART_IMMEDIATE_OFFLINE;
 		regs.bSectorNumberReg = select;
+		if (select == ABORT_SELF_TEST) // Abort only supported on Win9x, try
+			try_ioctl = 0x03; // SMART_SEND_DRIVE_COMMAND, then IOCTL_IDE_PASS_THROUGH
 		break;
 	  default:
 		pout("Unrecognized command %d in win32_ata_command_interface()\n"
@@ -742,27 +991,42 @@ int ata_command_interface(int fd, smart_command_set command, int select, char *
 		return -1;
 	}
 
-	if (smart_ioctl(h_ata_ioctl, fd, &regs, data, (copydata?512:0))) {
-		// Read log only supported on Win9x, retry with pass through command
-		// CAUTION: smart_ioctl() MUST NOT change "regs" Parameter in this case
-		if (errno == ENOSYS && command == READ_LOG && !ide_pass_through_broken) {
-			errno = 0;
-			memset(data, 0, 512);
-			if (ide_pass_through_ioctl(h_ata_ioctl, &regs, data, 512))
+	if (try_ioctl & 0x01) {
+		if (smart_ioctl(h_ata_ioctl, fd, &regs, data, (copydata?512:0))) {
+			if (!(try_ioctl & 0x02) || errno != ENOSYS)
 				return -1;
-			if (!nonempty(data, 512)) {
-				// Nothing useful returned => ioctl probably broken
-				pout("IOCTL_IDE_PASS_THROUGH does not work on your version of Windows\n");
-				ide_pass_through_broken = 1; // Do not retry (smartd)
+			// CAUTION: smart_ioctl() MUST NOT change "regs" Parameter in this case
+		}
+		else
+			try_ioctl = 0;
+	}
+
+	if (try_ioctl & 0x02) {
+		errno = 0;
+		if ((GetVersion() & 0x8000ffff) == 0x00000004) {
+			// Special case WinNT4
+			if (command == CHECK_POWER_MODE) { // SCSI_PASS_THROUGH does not return regs!
 				errno = ENOSYS;
 				return -1;
 			}
-			return 0;
+			if (ata_via_scsi_pass_through_ioctl(h_ata_ioctl, &regs, data, (copydata?512:0)))
+				return -1;
 		}
-		return -1;
+		else {
+			if (ide_pass_through_ioctl(h_ata_ioctl, &regs, data, (copydata?512:0)))
+				return -1;
+		}
+		try_ioctl = 0;
 	}
+	assert(!try_ioctl);
 
-	if (command == STATUS_CHECK) {
+	switch (command) {
+	  case CHECK_POWER_MODE:
+		// Return power mode from SectorCountReg in data[0]
+		data[0] = regs.bSectorCountReg;
+		return 0;
+
+	  case STATUS_CHECK:
 		// Cyl low and Cyl high unchanged means "Good SMART status"
 		if (regs.bCylHighReg == SMART_CYL_HI && regs.bCylLowReg == SMART_CYL_LOW)
 		  return 0;
@@ -777,28 +1041,60 @@ int ata_command_interface(int fd, smart_command_set command, int select, char *
 		print_ide_regs(&regs, 1);
 		errno = EIO;
 		return -1;
-	}
 
-	return 0;
+	  default:
+		return 0;
+	}
+	/*NOTREACHED*/
 }
 
 
+#ifndef HAVE_ATA_IDENTIFY_IS_CACHED
+#error define of HAVE_ATA_IDENTIFY_IS_CACHED missing in config.h
+#endif
+
+// Return true if OS caches the ATA identify sector
+int ata_identify_is_cached(int fd)
+{
+	ARGUSED(fd);
+	// WinNT4/2000/XP => true, Win9x/ME => false
+	return ((GetVersion() & 0x80000000) == 0);
+}
+
+
+// Print not implemeted warning once
+static void pr_not_impl(const char * what, int * warned)
+{
+	if (*warned)
+		return;
+	pout(
+		"#######################################################################\n"
+		"%s\n"
+		"NOT IMPLEMENTED under Win32.\n"
+		"Please contact " PACKAGE_BUGREPORT " if\n"
+		"you want to help in porting smartmontools to Win32.\n"
+		"#######################################################################\n"
+		"\n", what
+	);
+	*warned = 1;
+}
+
 // Interface to ATA devices behind 3ware escalade RAID controller cards.  See os_linux.c
 int escalade_command_interface(int fd, int disknum, int escalade_type, smart_command_set command, int select, char *data)
 {
 	static int warned = 0;
 	ARGUSED(fd); ARGUSED(disknum); ARGUSED(escalade_type); ARGUSED(command); ARGUSED(select); ARGUSED(data);
-	if (!warned) {
-		  pout(
-		"#######################################################################\n"
-		"3ware Escalade Controller command routine escalade_command_interface()\n"
-		"NOT IMPLEMENTED under Win32.\n"
-		"Please contact " PACKAGE_BUGREPORT " if\n"
-		"you want to help in porting smartmontools to Win32.\n"
-		"#######################################################################\n"
-		"\n");
-		warned = 1;
-	}
+	pr_not_impl("3ware Escalade Controller command routine escalade_command_interface()", &warned);
+	errno = ENOSYS;
+	return -1;
+}
+
+// Interface to ATA devices behind Marvell chip-set based controllers.  See os_linux.c
+int marvell_command_interface(int fd, smart_command_set command, int select, char * data)
+{
+	static int warned = 0;
+	ARGUSED(fd); ARGUSED(command); ARGUSED(select); ARGUSED(data);
+	pr_not_impl("Marvell chip-set command routine marvell_command_interface()", &warned);
 	errno = ENOSYS;
 	return -1;
 }
@@ -933,15 +1229,14 @@ static int aspi_call(ASPI_SRB * srb)
 	i = 0;
 	while (((volatile ASPI_SRB *)srb)->h.status == ASPI_STATUS_IN_PROGRESS) {
 		if (++i > 100/*10sek*/) {
-			pout("ASPI Adapter %u: Timeout\n", srb->h.adapter);
+			pout("ASPI Adapter %u: Timed out\n", srb->h.adapter);
 			aspi_entry = 0;
 			h_aspi_dll = INVALID_HANDLE_VALUE;
 			errno = EIO;
 			return -1;
 		}
-#ifdef _DEBUG
-		pout("ASPI Wait %d\n", i);
-#endif
+		if (con->reportscsiioctl > 1)
+			pout("ASPI Adapter %u: Waiting (%d) ...\n", srb->h.adapter, i);
 		Sleep(100);
 	}
 	return 0;
@@ -1001,6 +1296,13 @@ static int aspi_open_dll(int verbose)
 		errno = ENOENT;
 		return -1;
 	}
+	if (con->reportscsiioctl > 1) {
+		// Print full path of WNASPI32.DLL
+		char path[MAX_PATH];
+		if (!GetModuleFileName(h_aspi_dll, path, sizeof(path)))
+			strcpy(path, "*unknown*");
+		pout("Using ASPI interface \"%s\"\n", path);
+	}
 
 	// Get ASPI entrypoints
 	if (!(aspi_info = (UINT (*)(void))aspi_get_address("GetASPI32SupportInfo", verbose)))
@@ -1010,9 +1312,8 @@ static int aspi_open_dll(int verbose)
 
 	// Init ASPI manager and get number of adapters
 	info = (aspi_info)();
-#ifdef _DEBUG
-	pout("GetASPI32SupportInfo() returns 0x%04x\n", info);
-#endif
+	if (con->reportscsiioctl > 1)
+		pout("GetASPI32SupportInfo() returns 0x%04x\n", info);
 	rc = (info >> 8) & 0xff;
 	if (rc == ASPI_STATUS_NO_ADAPTERS) {
 		num_aspi_adapters = 0;
@@ -1030,6 +1331,9 @@ static int aspi_open_dll(int verbose)
 		return -1;
 	}
 
+	if (con->reportscsiioctl)
+		pout("%u ASPI Adapter%s detected\n",num_aspi_adapters, (num_aspi_adapters!=1?"s":""));
+
 #ifdef __CYGWIN__
 	// save PID to detect fork() in aspi_entry_valid()
 	aspi_dll_pid = GetCurrentProcessId();
@@ -1039,7 +1343,7 @@ static int aspi_open_dll(int verbose)
 }
 
 
-static int aspi_io_call(ASPI_SRB * srb)
+static int aspi_io_call(ASPI_SRB * srb, unsigned timeout)
 {
 	HANDLE event;
 	// Create event
@@ -1052,10 +1356,11 @@ static int aspi_io_call(ASPI_SRB * srb)
 	aspi_entry(srb);
 	if (((volatile ASPI_SRB *)srb)->h.status == ASPI_STATUS_IN_PROGRESS) {
 		// Wait for event
-		DWORD rc = WaitForSingleObject(event, 30*1000L);
+		DWORD rc = WaitForSingleObject(event, timeout*1000L);
 		if (rc != WAIT_OBJECT_0) {
 			if (rc == WAIT_TIMEOUT) {
-				pout("ASPI Timeout\n");
+				pout("ASPI Adapter %u, ID %u: Timed out after %u seconds\n",
+					srb->h.adapter, srb->i.target_id, timeout);
 			}
 			else {
 				pout("WaitForSingleObject(%lx) = 0x%lx,%ld, Error=%ld\n",
@@ -1074,6 +1379,7 @@ static int aspi_io_call(ASPI_SRB * srb)
 
 static int aspi_open(unsigned adapter, unsigned id)
 {
+	ASPI_SRB srb;
 	if (!(adapter <= 9 && id < 16)) {
 		errno = ENOENT;
 		return -1;
@@ -1086,13 +1392,32 @@ static int aspi_open(unsigned adapter, unsigned id)
 
 	// Adapter OK?
 	if (adapter >= num_aspi_adapters) {
-		pout("ASPI Adapter %d does not exist (%d Adapter(s) detected).\n", adapter, num_aspi_adapters);
+		pout("ASPI Adapter %u does not exist (%u Adapter%s detected).\n",
+			adapter, num_aspi_adapters, (num_aspi_adapters!=1?"s":""));
 		if (!is_permissive()) {
 			errno = ENOENT;
 			return -1;
 		}
 	}
 
+	// Device present ?
+	memset(&srb, 0, sizeof(srb));
+	srb.h.cmd = ASPI_CMD_GET_DEVICE_TYPE;
+	srb.h.adapter = adapter; srb.i.target_id = id;
+	if (aspi_call(&srb)) {
+		errno = EIO;
+		return -1;
+	}
+	if (srb.h.status != ASPI_STATUS_NO_ERROR) {
+		pout("ASPI Adapter %u, ID %u: No such device (Status=0x%02x)\n", adapter, id, srb.h.status);
+		if (!is_permissive()) {
+			errno = (srb.h.status == ASPI_STATUS_INVALID_TARGET ? ENOENT : EIO);
+			return -1;
+		}
+	}
+	else if (con->reportscsiioctl)
+		pout("ASPI Adapter %u, ID %u: Device Type=0x%02x\n", adapter, id, srb.t.devtype);
+
 	return (0x0100 | ((adapter & 0xf)<<4) | (id & 0xf));
 }
 
@@ -1104,39 +1429,48 @@ static void aspi_close(int fd)
 }
 
 
-// Scan for SCSI drives, return bitmask [adapter:0-3][id:0-7] of drives present
+// Scan for SCSI drives, fill bitmask [adapter:0-9][id:0-7] of drives present,
+// return #drives
 
-static unsigned long aspi_scan()
+static int aspi_scan(unsigned long * drives)
 {
-	unsigned long drives = 0;
-	unsigned ad, nad;
+	int cnt = 0;
+	unsigned ad;
 
 	if (!aspi_entry_valid()) {
-		if (aspi_open_dll(0/*quiet*/))
+		if (aspi_open_dll(con->reportscsiioctl/*default is quiet*/))
 			return 0;
 	}
 
-	nad = num_aspi_adapters;
-	if (nad >= 4)
-		nad = 4;
-	for (ad = 0; ad < nad; ad++) {
-		ASPI_SRB srb; int id;
+	for (ad = 0; ad < num_aspi_adapters; ad++) {
+		ASPI_SRB srb; unsigned id;
+
+		if (ad > 9) {
+			if (con->reportscsiioctl)
+				pout(" ASPI Adapter %u: Ignored\n", ad);
+			continue;
+		}
+
 		// Get adapter name
 		memset(&srb, 0, sizeof(srb));
 		srb.h.cmd = ASPI_CMD_ADAPTER_INQUIRE;
 		srb.h.adapter = ad;
 		if (aspi_call(&srb))
 			return 0;
-#ifdef _DEBUG
-		pout("ASPI Adapter %u: %02x,\"%.16s\"\n", ad, srb.h.status, srb.q.adapter_id);
-#endif
-		if (srb.h.status != ASPI_STATUS_NO_ERROR)
-			continue;
 
-		// Skip ATA/ATAPI devices
-		srb.q.adapter_id[sizeof(srb.q.adapter_id)-1] = 0;
-		if (strstr(srb.q.adapter_id, "ATAPI"))
+		if (srb.h.status != ASPI_STATUS_NO_ERROR) {
+			if (con->reportscsiioctl)
+				pout(" ASPI Adapter %u: Status=0x%02x\n", ad, srb.h.status);
 			continue;
+		}
+
+		if (con->reportscsiioctl) {
+			int i;
+			for (i = 1; i < 16 && srb.q.adapter_id[i]; i++)
+				if (!(' ' <= srb.q.adapter_id[i] && srb.q.adapter_id[i] <= '~'))
+					srb.q.adapter_id[i] = '?';
+			pout(" ASPI Adapter %u (\"%.16s\"):\n", ad, srb.q.adapter_id);
+		}
 
 		for (id = 0; id <= 7; id++) {
 			// Get device type
@@ -1145,14 +1479,20 @@ static unsigned long aspi_scan()
 			srb.h.adapter = ad; srb.i.target_id = id;
 			if (aspi_call(&srb))
 				return 0;
-#ifdef _DEBUG
-			pout("Device type for scsi%u%x: %02x,%02x\n", ad, id, srb.h.status, srb.t.devtype);
-#endif
-			if (srb.h.status == ASPI_STATUS_NO_ERROR && srb.t.devtype == 0x00/*HDD*/)
-				drives |= 1 << ((ad<<3)+id);
+			if (srb.h.status != ASPI_STATUS_NO_ERROR) {
+				if (con->reportscsiioctl > 1)
+					pout("  ID %u: No such device (Status=0x%02x)\n", id, srb.h.status);
+				continue;
+			}
+			if (con->reportscsiioctl)
+				pout("  ID %u: Device Type=0x%02x\n", id, srb.t.devtype);
+			if (srb.t.devtype == 0x00/*HDD*/) {
+				drives[ad >> 2] |= (1L << (((ad & 0x3) << 3) + id));
+				cnt++;
+			}
 		}
 	}
-	return drives;
+	return cnt;
 }
 
 
@@ -1231,7 +1571,7 @@ int do_scsi_cmnd_io(int fd, struct scsi_cmnd_io * iop, int report)
 	iop->scsi_status = 0;
 	iop->resid = 0;
 
-	if (aspi_io_call(&srb)) {
+	if (aspi_io_call(&srb, (iop->timeout ? iop->timeout : 60))) {
 		// Timeout
 		return -EIO;
 	}
