@@ -3,11 +3,11 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 1999-2000 Michael Cornwell <cornwell@acm.org>
  *
  * Additional SCSI work:
- * Copyright (C) 2003-4 Douglas Gilbert <dougg@torque.net>
+ * Copyright (C) 2003-6 Douglas Gilbert <dougg@torque.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -47,7 +47,7 @@
 #include "scsicmds.h"
 #include "utility.h"
 
-const char *scsicmds_c_cvsid="$Id: scsicmds.c,v 1.78 2004/09/05 13:53:14 dpgilbert Exp $"
+const char *scsicmds_c_cvsid="$Id: scsicmds.c,v 1.85 2006/04/12 14:54:28 ballen4705 Exp $"
 CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID SCSICMDS_H_CVSID UTILITY_H_CVSID;
 
 /* for passing global control variables */
@@ -120,6 +120,7 @@ static struct scsi_opcode_name opcode_name_arr[] = {
     {MODE_SENSE, "mode sense"},                 /* 0x1a */
     {RECEIVE_DIAGNOSTIC, "receive diagnostic"}, /* 0x1c */
     {SEND_DIAGNOSTIC, "send diagnostic"},       /* 0x1d */
+    {READ_DEFECT_10, "read defect list(10)"},   /* 0x37 */
     {LOG_SENSE, "log sense"},                   /* 0x4d */
     {MODE_SELECT_10, "mode select(10)"},        /* 0x55 */
     {MODE_SENSE_10, "mode sense(10)"},          /* 0x5a */
@@ -159,7 +160,8 @@ void scsi_do_sense_disect(const struct scsi_cmnd_io * io_buf,
 
 static int scsiSimpleSenseFilter(const struct scsi_sense_disect * sinfo)
 {
-    if (SCSI_SK_NOT_READY == sinfo->sense_key) {
+    switch (sinfo->sense_key) {
+    case SCSI_SK_NOT_READY:
         if (SCSI_ASC_NO_MEDIUM == sinfo->asc) 
             return SIMPLE_ERR_NO_MEDIUM;
         else if (SCSI_ASC_NOT_READY == sinfo->asc) {
@@ -169,16 +171,23 @@ static int scsiSimpleSenseFilter(const struct scsi_sense_disect * sinfo)
                 return SIMPLE_ERR_NOT_READY;
         } else
             return SIMPLE_ERR_NOT_READY;
-    } else if (SCSI_SK_ILLEGAL_REQUEST == sinfo->sense_key) {
+    case SCSI_SK_MEDIUM_ERROR:
+    case SCSI_SK_HARDWARE_ERROR:
+        return SIMPLE_ERR_MEDIUM_HARDWARE;
+    case SCSI_SK_ILLEGAL_REQUEST:
         if (SCSI_ASC_UNKNOWN_OPCODE == sinfo->asc)
             return SIMPLE_ERR_BAD_OPCODE;
         else if (SCSI_ASC_UNKNOWN_FIELD == sinfo->asc)
             return SIMPLE_ERR_BAD_FIELD;
         else if (SCSI_ASC_UNKNOWN_PARAM == sinfo->asc)
             return SIMPLE_ERR_BAD_PARAM;
-    } else if (SCSI_SK_UNIT_ATTENTION == sinfo->sense_key)
+        else
+            return SIMPLE_ERR_BAD_PARAM;    /* all other illegal request */
+    case SCSI_SK_UNIT_ATTENTION:
         return SIMPLE_ERR_TRY_AGAIN;
-    return SIMPLE_NO_ERROR;
+    default:
+        return SIMPLE_NO_ERROR;
+    }
 }
 
 const char * scsiErrString(int scsiErr)
@@ -204,6 +213,8 @@ const char * scsiErrString(int scsiErr)
             return "device will be ready soon";
         case SIMPLE_ERR_TRY_AGAIN: 
             return "unit attention reported, try again";
+        case SIMPLE_ERR_MEDIUM_HARDWARE: 
+            return "medium or hardware error (serious)";
         default:
             return "unknown error";
     }
@@ -211,8 +222,8 @@ const char * scsiErrString(int scsiErr)
 
 /* Sends LOG SENSE command. Returns 0 if ok, 1 if device NOT READY, 2 if
    command not supported, 3 if field (within command) not supported or
-   returns negated errno.  SPC sections 7.6 and 8.2 N.B. Sets PC==1
-   to fetch "current cumulative" log pages.
+   returns negated errno.  SPC-3 sections 6.6 and 7.2 (rec 22a).
+   N.B. Sets PC==1 to fetch "current cumulative" log pages.
    If known_resp_len > 0 then a single fetch is done for this response
    length. If known_resp_len == 0 then twin fetches are performed, the
    first to deduce the response length, then send the same command again
@@ -308,7 +319,7 @@ int scsiLogSense(int device, int pagenum, UINT8 *pBuf, int bufLen,
 /* Send MODE SENSE (6 byte) command. Returns 0 if ok, 1 if NOT READY,
  * 2 if command not supported (then MODE SENSE(10) should be supported),
  * 3 if field in command not supported or returns negated errno. 
- * SPC sections 7.9 and 8.4 [mode subpage==0] */
+ * SPC-3 sections 6.9 and 7.4 (rev 22a) [mode subpage==0] */
 int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -335,13 +346,13 @@ int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
 
     status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
-	return status;
+        return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
     status = scsiSimpleSenseFilter(&sinfo);
     if (SIMPLE_ERR_TRY_AGAIN == status) {
         status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
         if (0 != status)
-	    return status;
+            return status;
         scsi_do_sense_disect(&io_hdr, &sinfo);
         status = scsiSimpleSenseFilter(&sinfo);
     }
@@ -363,7 +374,7 @@ int scsiModeSense(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
  * (normally 1) and then 1 mode page. Returns 0 if ok, 1 if NOT READY,
  * 2 if command not supported (then MODE SELECT(10) may be supported), 
  * 3 if field in command not supported, 4 if bad parameter to command
- * or returns negated errno. SPC sections 7.7 and 8.4 */
+ * or returns negated errno. SPC-3 sections 6.7 and 7.4 (rev 22a) */
 int scsiModeSelect(int device, int sp, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -380,7 +391,7 @@ int scsiModeSelect(int device, int sp, UINT8 *pBuf, int bufLen)
     if (hdr_plus_1_pg > bufLen)
         return -EINVAL;
     pBuf[0] = 0;    /* Length of returned mode sense data reserved for SELECT */
-    pBuf[pg_offset] &= 0x3f;    /* Mask of PS bit from byte 0 of page data */
+    pBuf[pg_offset] &= 0x7f;    /* Mask out PS bit from byte 0 of page data */
     memset(&io_hdr, 0, sizeof(io_hdr));
     memset(cdb, 0, sizeof(cdb));
     io_hdr.dxfer_dir = DXFER_TO_DEVICE;
@@ -405,7 +416,7 @@ int scsiModeSelect(int device, int sp, UINT8 *pBuf, int bufLen)
 /* MODE SENSE (10 byte). Returns 0 if ok, 1 if NOT READY, 2 if command 
  * not supported (then MODE SENSE(6) might be supported), 3 if field in
  * command not supported or returns negated errno.  
- * SPC sections 7.10 and 8.4 [mode subpage==0] */
+ * SPC-3 sections 6.10 and 7.4 (rev 22a) [mode subpage==0] */
 int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -431,13 +442,13 @@ int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
 
     status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
-	return status;
+        return status;
     scsi_do_sense_disect(&io_hdr, &sinfo);
     status = scsiSimpleSenseFilter(&sinfo);
     if (SIMPLE_ERR_TRY_AGAIN == status) {
         status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
         if (0 != status)
-	    return status;
+            return status;
         scsi_do_sense_disect(&io_hdr, &sinfo);
         status = scsiSimpleSenseFilter(&sinfo);
     }
@@ -459,7 +470,7 @@ int scsiModeSense10(int device, int pagenum, int pc, UINT8 *pBuf, int bufLen)
  * (normally 1) and then 1 mode page. Returns 0 if ok, 1 NOT REAFY, 2 if 
  * command not supported (then MODE SELECT(6) may be supported), 3 if field
  * in command not supported, 4 if bad parameter to command or returns
- * negated errno. SAM sections 7.8 and 8.4 */
+ * negated errno. SPC-3 sections 6.8 and 7.4 (rev 22a) */
 int scsiModeSelect10(int device, int sp, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -477,7 +488,7 @@ int scsiModeSelect10(int device, int sp, UINT8 *pBuf, int bufLen)
         return -EINVAL;
     pBuf[0] = 0;    
     pBuf[1] = 0; /* Length of returned mode sense data reserved for SELECT */
-    pBuf[pg_offset] &= 0x3f;    /* Mask of PS bit from byte 0 of page data */
+    pBuf[pg_offset] &= 0x7f;    /* Mask out PS bit from byte 0 of page data */
     memset(&io_hdr, 0, sizeof(io_hdr));
     memset(cdb, 0, sizeof(cdb));
     io_hdr.dxfer_dir = DXFER_TO_DEVICE;
@@ -501,12 +512,13 @@ int scsiModeSelect10(int device, int sp, UINT8 *pBuf, int bufLen)
 
 /* Standard INQUIRY returns 0 for ok, anything else is a major problem.
  * bufLen should be 36 for unsafe devices (like USB mass storage stuff)
- * otherwise they can lock up! SPC sections 7.4 and 8.6 */
+ * otherwise they can lock up! SPC-3 sections 6.4 and 7.6 (rev 22a) */
 int scsiStdInquiry(int device, UINT8 *pBuf, int bufLen)
 {
     struct scsi_sense_disect sinfo;
     struct scsi_cmnd_io io_hdr;
     UINT8 cdb[6];
+    UINT8 sense[32];
     int status;
 
     if ((bufLen < 0) || (bufLen > 255))
@@ -520,6 +532,10 @@ int scsiStdInquiry(int device, UINT8 *pBuf, int bufLen)
     cdb[4] = bufLen;
     io_hdr.cmnd = cdb;
     io_hdr.cmnd_len = sizeof(cdb);
+    io_hdr.sensep = sense;
+    io_hdr.max_sense_len = sizeof(sense);
+    io_hdr.timeout = SCSI_TIMEOUT_DEFAULT;
+
     status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
         return status;
@@ -530,7 +546,7 @@ int scsiStdInquiry(int device, UINT8 *pBuf, int bufLen)
 /* INQUIRY to fetch Vital Page Data.  Returns 0 if ok, 1 if NOT READY
  * (unlikely), 2 if command not supported, 3 if field in command not 
  * supported, 5 if response indicates that EVPD bit ignored or returns
- * negated errno. SPC section 7.4 and 8.6 */
+ * negated errno. SPC-3 section 6.4 and 7.6 (rev 22a) */
 int scsiInquiryVpd(int device, int vpd_page, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -576,11 +592,12 @@ int scsiInquiryVpd(int device, int vpd_page, UINT8 *pBuf, int bufLen)
 }
 
 /* REQUEST SENSE command. Returns 0 if ok, anything else major problem.
- * SPC section 7.24 */
+ * SPC-3 section 6.27 (rev 22a) */
 int scsiRequestSense(int device, struct scsi_sense_disect * sense_info)
 {
     struct scsi_cmnd_io io_hdr;
     UINT8 cdb[6];
+    UINT8 sense[32];
     UINT8 buff[18];
     int status, len;
     UINT8 ecode;
@@ -594,6 +611,10 @@ int scsiRequestSense(int device, struct scsi_sense_disect * sense_info)
     cdb[4] = sizeof(buff);
     io_hdr.cmnd = cdb;
     io_hdr.cmnd_len = sizeof(cdb);
+    io_hdr.sensep = sense;
+    io_hdr.max_sense_len = sizeof(sense);
+    io_hdr.timeout = SCSI_TIMEOUT_DEFAULT;
+
     status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if ((0 == status) && (sense_info)) {
         ecode = buff[0] & 0x7f;
@@ -614,7 +635,7 @@ int scsiRequestSense(int device, struct scsi_sense_disect * sense_info)
 
 /* SEND DIAGNOSTIC command.  Returns 0 if ok, 1 if NOT READY, 2 if command
  * not supported, 3 if field in command not supported or returns negated
- * errno. SPC section 7.25 */
+ * errno. SPC-3 section 6.28 (rev 22a) */
 int scsiSendDiagnostic(int device, int functioncode, UINT8 *pBuf, int bufLen)
 {
     struct scsi_cmnd_io io_hdr;
@@ -641,8 +662,8 @@ int scsiSendDiagnostic(int device, int functioncode, UINT8 *pBuf, int bufLen)
     io_hdr.cmnd_len = sizeof(cdb);
     io_hdr.sensep = sense;
     io_hdr.max_sense_len = sizeof(sense);
-    io_hdr.timeout = SCSI_TIMEOUT_SELF_TEST;
     /* worst case is an extended foreground self test on a big disk */
+    io_hdr.timeout = SCSI_TIMEOUT_SELF_TEST;
     
     status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
     if (0 != status)
@@ -653,7 +674,7 @@ int scsiSendDiagnostic(int device, int functioncode, UINT8 *pBuf, int bufLen)
 
 /* RECEIVE DIAGNOSTIC command. Returns 0 if ok, 1 if NOT READY, 2 if
  * command not supported, 3 if field in command not supported or returns
- * negated errno. SPC section 7.17 */
+ * negated errno. SPC-3 section 6.18 (rev 22a) */
 int scsiReceiveDiagnostic(int device, int pcv, int pagenum, UINT8 *pBuf, 
                       int bufLen)
 {
@@ -686,7 +707,7 @@ int scsiReceiveDiagnostic(int device, int pcv, int pagenum, UINT8 *pBuf,
     return scsiSimpleSenseFilter(&sinfo);
 }
 
-/* TEST UNIT READY command. SPC section 7.28 (probably in SBC as well) */
+/* TEST UNIT READY command. SPC-3 section 6.33 (rev 22a) */
 static int _testunitready(int device, struct scsi_sense_disect * sinfo)
 {
     struct scsi_cmnd_io io_hdr;
@@ -734,6 +755,41 @@ int scsiTestUnitReady(int device)
     return status;
 }
 
+/* READ DEFECT (10) command. Returns 0 if ok, 1 if NOT READY, 2 if
+ * command not supported, 3 if field in command not supported or returns
+ * negated errno. SBC-2 section 5.12 (rev 16) */
+int scsiReadDefect10(int device, int req_plist, int req_glist, int dl_format,
+                     UINT8 *pBuf, int bufLen)
+{
+    struct scsi_cmnd_io io_hdr;
+    struct scsi_sense_disect sinfo;
+    UINT8 cdb[10];
+    UINT8 sense[32];
+    int status;
+
+    memset(&io_hdr, 0, sizeof(io_hdr));
+    memset(cdb, 0, sizeof(cdb));
+    io_hdr.dxfer_dir = DXFER_FROM_DEVICE;
+    io_hdr.dxfer_len = bufLen;
+    io_hdr.dxferp = pBuf;
+    cdb[0] = READ_DEFECT_10;
+    cdb[2] = (unsigned char)(((req_plist << 4) & 0x10) |
+               ((req_glist << 3) & 0x8) | (dl_format & 0x7));
+    cdb[7] = (bufLen >> 8) & 0xff;
+    cdb[8] = bufLen & 0xff;
+    io_hdr.cmnd = cdb;
+    io_hdr.cmnd_len = sizeof(cdb);
+    io_hdr.sensep = sense;
+    io_hdr.max_sense_len = sizeof(sense);
+    io_hdr.timeout = SCSI_TIMEOUT_DEFAULT;
+
+    status = do_scsi_cmnd_io(device, &io_hdr, con->reportscsiioctl);
+    if (0 != status)
+        return status;
+    scsi_do_sense_disect(&io_hdr, &sinfo);
+    return scsiSimpleSenseFilter(&sinfo);
+}
+
 /* Offset into mode sense (6 or 10 byte) response that actual mode page
  * starts at (relative to resp[0]). Returns -1 if problem */
 int scsiModePageOffset(const UINT8 * resp, int len, int modese_len)
@@ -756,8 +812,10 @@ int scsiModePageOffset(const UINT8 * resp, int len, int modese_len)
                  "resp_len=%d bd_len=%d\n", offset, resp_len, bd_len);
             offset = -1;
         } else if ((offset + 2) > resp_len) {
-            pout("scsiModePageOffset: response length too short, resp_len=%d"
-                 " offset=%d bd_len=%d\n", resp_len, offset, bd_len);
+             if ((resp_len > 2) || con->reportscsiioctl)
+                pout("scsiModePageOffset: response length too short, "
+                     "resp_len=%d offset=%d bd_len=%d\n", resp_len,
+                     offset, bd_len);
             offset = -1;
         }
     }
@@ -882,14 +940,13 @@ int scsiSetExceptionControlAndWarning(int device, int enabled,
         return -EINVAL;
     memcpy(rout, iecp->raw_curr, SCSI_IECMP_RAW_LEN);
     if (10 == iecp->modese_len) {
-            resp_len = (rout[0] << 8) + rout[1] + 2;
-            memset(rout, 0, 2); /* mode data length==0 for mode select */
+        resp_len = (rout[0] << 8) + rout[1] + 2;
+        rout[3] &= 0xef;    /* for disks mask out DPOFUA bit */
     } else {
-            resp_len = rout[0] + 1;
-            memset(rout, 0, 1); /* mode data length==0 for mode select */
+        resp_len = rout[0] + 1;
+        rout[2] &= 0xef;    /* for disks mask out DPOFUA bit */
     }
     sp = (rout[offset] & 0x80) ? 1 : 0; /* PS bit becomes 'SELECT's SP bit */
-    rout[offset] &= 0x7f;     /* mask off PS bit */
     if (enabled) {
         rout[offset + 2] = SCSI_IEC_MP_BYTE2_ENABLED;
         if (con->reportscsiioctl > 2)
@@ -1550,6 +1607,7 @@ const char * scsiGetIEString(UINT8 asc, UINT8 ascq)
 int scsiSmartIBMOfflineTest(int device)
 {       
     UINT8 tBuf[256];
+    int res;
         
     memset(tBuf, 0, sizeof(tBuf));
     /* Build SMART Off-line Immediate Diag Header */
@@ -1561,39 +1619,72 @@ int scsiSmartIBMOfflineTest(int device)
     tBuf[5] = 0x00; /* Reserved */
     tBuf[6] = 0x00; /* Off-line Immediate Time MSB */
     tBuf[7] = 0x00; /* Off-line Immediate Time LSB */
-    return scsiSendDiagnostic(device, SCSI_DIAG_NO_SELF_TEST, tBuf, 8);
+    res = scsiSendDiagnostic(device, SCSI_DIAG_NO_SELF_TEST, tBuf, 8);
+    if (res)
+        pout("IBM offline test failed [%s]\n", scsiErrString(res));
+    return res;
 }
 
 int scsiSmartDefaultSelfTest(int device)
 {       
-    return scsiSendDiagnostic(device, SCSI_DIAG_DEF_SELF_TEST, NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_DEF_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Default self test failed [%s]\n", scsiErrString(res));
+    return res;
 }
 
 int scsiSmartShortSelfTest(int device)
 {       
-    return scsiSendDiagnostic(device, SCSI_DIAG_BG_SHORT_SELF_TEST, NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_BG_SHORT_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Short offline self test failed [%s]\n", scsiErrString(res));
+    return res;
 }
 
 int scsiSmartExtendSelfTest(int device)
 {       
-    return scsiSendDiagnostic(device, SCSI_DIAG_BG_EXTENDED_SELF_TEST, 
-                              NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_BG_EXTENDED_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Long (extended) offline self test failed [%s]\n",
+             scsiErrString(res));
+    return res;
 }
 
 int scsiSmartShortCapSelfTest(int device)
 {       
-    return scsiSendDiagnostic(device, SCSI_DIAG_FG_SHORT_SELF_TEST, NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_FG_SHORT_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Short foreground self test failed [%s]\n", scsiErrString(res));
+    return res;
 }
 
 int scsiSmartExtendCapSelfTest(int device)
 {
-    return scsiSendDiagnostic(device, SCSI_DIAG_FG_EXTENDED_SELF_TEST, 
-                              NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_FG_EXTENDED_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Long (extended) foreground self test failed [%s]\n",
+             scsiErrString(res));
+    return res;
 }
 
 int scsiSmartSelfTestAbort(int device)
 {
-    return scsiSendDiagnostic(device, SCSI_DIAG_ABORT_SELF_TEST, NULL, 0);
+    int res;
+
+    res = scsiSendDiagnostic(device, SCSI_DIAG_ABORT_SELF_TEST, NULL, 0);
+    if (res)
+        pout("Abort self test failed [%s]\n", scsiErrString(res));
+    return res;
 }
 
 /* Returns 0 and the expected duration of an extended self test (in seconds)
@@ -1757,7 +1848,7 @@ void scsiDecodeNonMediumErrPage(unsigned char *resp,
    are the poweron hour of the most recent failure. Note: aborted self
    tests (typically by the user) and self tests in progress are not 
    considered failures. See Working Draft SCSI Primary Commands - 3 
-   (SPC-3) section 7.2.10 T10/1416-D Rev 15 */
+   (SPC-3) section 7.2.10 T10/1416-D (rev 22a) */
 int scsiCountFailedSelfTests(int fd, int noisy)
 {
     int num, k, n, err, res, fails, fail_hour;
@@ -1915,14 +2006,13 @@ int scsiSetControlGLTSD(int device, int enabled, int modese_len)
         return SIMPLE_ERR_BAD_PARAM;  /* GLTSD bit not chageable */
     
     if (10 == modese_len) {
-            resp_len = (buff[0] << 8) + buff[1] + 2;
-            memset(buff, 0, 2);
+        resp_len = (buff[0] << 8) + buff[1] + 2;
+        buff[3] &= 0xef;    /* for disks mask out DPOFUA bit */
     } else {
-            resp_len = buff[0] + 1;
-            memset(buff, 0, 1);
+        resp_len = buff[0] + 1;
+        buff[2] &= 0xef;    /* for disks mask out DPOFUA bit */
     }
     sp = (buff[offset] & 0x80) ? 1 : 0; /* PS bit becomes 'SELECT's SP bit */
-    buff[offset] &= 0x7f;     /* mask off PS bit */
     if (enabled)
         buff[offset + 2] |= 0x2;    /* set GLTSD bit */
     else
