@@ -3,11 +3,11 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * Additional SCSI work:
- * Copyright (C) 2003-4 Douglas Gilbert <dougg@torque.net>
+ * Copyright (C) 2003-6 Douglas Gilbert <dougg@torque.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,15 +32,13 @@
 #ifndef SCSICMDS_H_
 #define SCSICMDS_H_
 
-#define SCSICMDS_H_CVSID "$Id: scsicmds.h,v 1.52 2004/09/05 13:52:05 dpgilbert Exp $\n"
+#define SCSICMDS_H_CVSID "$Id: scsicmds.h,v 1.57 2006/04/12 14:54:28 ballen4705 Exp $\n"
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 
-#include "int64.h"
-
 /* #define SCSI_DEBUG 1 */ /* Comment out to disable command debugging */
 
 /* Following conditional defines bypass inclusion of scsi/scsi.h and
@@ -75,6 +73,9 @@
 #ifndef SEND_DIAGNOSTIC
 #define SEND_DIAGNOSTIC  0x1d
 #endif
+#ifndef READ_DEFECT_10
+#define READ_DEFECT_10  0x37
+#endif
 
 typedef unsigned char UINT8;
 typedef char INT8;
@@ -212,6 +213,8 @@ Documentation, see http://www.storage.ibm.com/techsup/hddtech/prodspecs.htm */
 
 /* defines for useful Sense Key codes */
 #define SCSI_SK_NOT_READY               0x2
+#define SCSI_SK_MEDIUM_ERROR            0x3
+#define SCSI_SK_HARDWARE_ERROR          0x4
 #define SCSI_SK_ILLEGAL_REQUEST         0x5
 #define SCSI_SK_UNIT_ATTENTION          0x6
 
@@ -234,6 +237,7 @@ Documentation, see http://www.storage.ibm.com/techsup/hddtech/prodspecs.htm */
 #define SIMPLE_ERR_NO_MEDIUM            6       /* no medium present */
 #define SIMPLE_ERR_BECOMING_READY       7       /* device will be ready soon */
 #define SIMPLE_ERR_TRY_AGAIN            8       /* some warning, try again */
+#define SIMPLE_ERR_MEDIUM_HARDWARE      9       /* medium or hardware error */
 
 
 /* defines for functioncode parameter in SENDDIAGNOSTIC function */
@@ -287,6 +291,9 @@ int scsiSendDiagnostic(int device, int functioncode, UINT8 *pBuf, int bufLen);
 int scsiReceiveDiagnostic(int device, int pcv, int pagenum, UINT8 *pBuf,
                       int bufLen);
 
+int scsiReadDefect10(int device, int req_plist, int req_glist, int dl_format,
+                     UINT8 *pBuf, int bufLen);
+
 /* SMART specific commands */
 int scsiCheckIE(int device, int hasIELogPage, int hasTempLogPage, UINT8 *asc,
                 UINT8 *ascq, UINT8 *currenttemp, UINT8 *triptemp);
