@@ -3,11 +3,11 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * Additional SCSI work:
- * Copyright (C) 2003-4 Douglas Gilbert <dougg@torque.net>
+ * Copyright (C) 2003-6 Douglas Gilbert <dougg@torque.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,7 +41,7 @@
 
 #define GBUF_SIZE 65535
 
-const char* scsiprint_c_cvsid="$Id: scsiprint.c,v 1.89 2004/09/06 03:41:13 dpgilbert Exp $"
+const char* scsiprint_c_cvsid="$Id: scsiprint.c,v 1.107 2006/04/12 16:18:57 ballen4705 Exp $"
 CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // control block which points to external global control variables
@@ -60,6 +60,11 @@ static int gSmartLPage = 0;     /* Informational Exceptions log page */
 static int gTempLPage = 0;
 static int gSelfTestLPage = 0;
 static int gStartStopLPage = 0;
+static int gReadECounterLPage = 0;
+static int gWriteECounterLPage = 0;
+static int gVerifyECounterLPage = 0;
+static int gNonMediumELPage = 0;
+static int gLastNErrorLPage = 0;
 static int gTapeAlertsLPage = 0;
 static int gSeagateCacheLPage = 0;
 static int gSeagateFactoryLPage = 0;
@@ -89,6 +94,21 @@ static void scsiGetSupportedLogPages(int device)
     for (i = 4; i < gBuf[3] + LOGPAGEHDRSIZE; i++) {
         switch (gBuf[i])
         {
+            case READ_ERROR_COUNTER_LPAGE:
+                gReadECounterLPage = 1;
+                break;
+            case WRITE_ERROR_COUNTER_LPAGE:
+                gWriteECounterLPage = 1;
+                break;
+            case VERIFY_ERROR_COUNTER_LPAGE:
+                gVerifyECounterLPage = 1;
+                break;
+            case LAST_N_ERROR_LPAGE:
+                gLastNErrorLPage = 1;
+                break;
+            case NON_MEDIUM_ERROR_LPAGE:
+                gNonMediumELPage = 1;
+                break;
             case TEMPERATURE_LPAGE:
                 gTempLPage = 1;
                 break;
@@ -116,27 +136,30 @@ static void scsiGetSupportedLogPages(int device)
     }
 }
 
-void scsiGetSmartData(int device, int attribs)
+/* Returns 0 if ok, -1 if can't check IE, -2 if can check and bad
+   (or at least something to report). */
+static int scsiGetSmartData(int device, int attribs)
 {
     UINT8 asc;
     UINT8 ascq;
     UINT8 currenttemp = 0;
     UINT8 triptemp = 0;
     const char * cp;
-    int err;
+    int err = 0;
 
     PRINT_ON(con);
-    if ((err = scsiCheckIE(device, gSmartLPage, gTempLPage, &asc, 
-                           &ascq, &currenttemp, &triptemp))) {
+    if (scsiCheckIE(device, gSmartLPage, gTempLPage, &asc, &ascq,
+                    &currenttemp, &triptemp)) {
         /* error message already announced */
         PRINT_OFF(con);
-        return;
+        return -1;
     }
     PRINT_OFF(con);
     cp = scsiGetIEString(asc, ascq);
     if (cp) {
+        err = -2;
         PRINT_ON(con);
-        pout("SMART Health Status: %s [asc=%x,ascq=%x]\n", cp, asc, ascq); 
+        pout("SMART Health Status: %s [asc=%x, ascq=%x]\n", cp, asc, ascq); 
         PRINT_OFF(con);
     } else if (gIecMPage)
         pout("SMART Health Status: OK\n");
@@ -153,6 +176,7 @@ void scsiGetSmartData(int device, int attribs)
         if (triptemp)
             pout("Drive Trip Temperature:        %d C\n", triptemp);
     }
+    return err;
 }
 
 
@@ -208,7 +232,7 @@ static int scsiGetTapeAlertsData(int device, int peripheral_type)
     return failures;
 }
 
-void scsiGetStartStopData(int device)
+static void scsiGetStartStopData(int device)
 {
     UINT32 currentStartStop;
     UINT32 recommendedStartStop; 
@@ -245,11 +269,61 @@ void scsiGetStartStopData(int device)
         currentStartStop = (gBuf[36] << 24) | (gBuf[37] << 16) |
                            (gBuf[38] << 8) | gBuf[39];
         pout("Current start stop count:      %u times\n", currentStartStop);
-        pout("Recommended maximum start stop count:  %u times\n", 
-             recommendedStartStop);
+	if (0xffffffff != recommendedStartStop)
+            pout("Recommended maximum start stop count:  %u times\n", 
+                 recommendedStartStop);
     }
 } 
 
+static void scsiPrintGrownDefectListLen(int device)
+{
+    int err, dl_format, dl_len, div;
+
+    memset(gBuf, 0, 4);
+    if ((err = scsiReadDefect10(device, 0 /* req_plist */, 1 /* req_glist */,
+                                4 /* bytes from index */, gBuf, 4))) {
+        if (con->reportscsiioctl > 0) {
+            PRINT_ON(con);
+            pout("Read defect list (10) Failed: %s\n", scsiErrString(err));
+            PRINT_OFF(con);
+        }
+        return;
+    }
+    if (0x8 != (gBuf[1] & 0x18)) {
+        PRINT_ON(con);
+        pout("Read defect list: asked for grown list but didn't get it\n");
+        PRINT_OFF(con);
+        return;
+    }
+    div = 0;
+    dl_format = (gBuf[1] & 0x7);
+    switch (dl_format) {
+        case 0:     /* short block */
+            div = 4;
+            break;
+        case 3:     /* long block */
+        case 4:     /* bytes from index */
+        case 5:     /* physical sector */
+            div = 8;
+            break;
+        default:
+            PRINT_ON(con);
+            pout("defect list format %d unknown\n", dl_format);
+            PRINT_OFF(con);
+            break;
+    }
+    dl_len = (gBuf[2] << 8) + gBuf[3];
+    if (0 == dl_len)
+        pout("Elements in grown defect list: 0\n");
+    else {
+        if (0 == div)
+            pout("Grown defect list length=%d bytes [unknown "
+                 "number of elements]\n", dl_len);
+        else
+            pout("Elements in grown defect list: %d\n", dl_len / div);
+    }
+}
+
 static void scsiPrintSeagateCacheLPage(int device)
 {
     int k, j, num, pl, pc, err, len;
@@ -327,7 +401,7 @@ static void scsiPrintSeagateCacheLPage(int device)
 
 static void scsiPrintSeagateFactoryLPage(int device)
 {
-    int k, j, num, pl, pc, len, err;
+    int k, j, num, pl, pc, len, err, good, bad;
     unsigned char * ucp;
     unsigned char * xp;
     uint64_t ull;
@@ -348,52 +422,72 @@ static void scsiPrintSeagateFactoryLPage(int device)
     len = ((gBuf[2] << 8) | gBuf[3]) + 4;
     num = len - 4;
     ucp = &gBuf[0] + 4;
+    good = 0;
+    bad = 0;
     while (num > 3) {
         pc = (ucp[0] << 8) | ucp[1];
         pl = ucp[3] + 4;
         switch (pc) {
         case 0: case 8:
+            ++good;
             break;
         default: 
-            if (con->reportscsiioctl > 0) {
-                PRINT_ON(con);
-                pout("\nVendor (Seagate/Hitachi) factory lpage has unexpected"
-                     " parameter, skip\n");
-                PRINT_OFF(con);
-            }
-            return;
+            ++bad;
+            break;
         }
         num -= pl;
         ucp += pl;
     }
+    if ((good < 2) || (bad > 4)) {  /* heuristic */
+        if (con->reportscsiioctl > 0) {
+            PRINT_ON(con);
+            pout("\nVendor (Seagate/Hitachi) factory lpage has too many "
+                 "unexpected parameters, skip\n");
+            PRINT_OFF(con);
+        }
+        return;
+    }
     pout("Vendor (Seagate/Hitachi) factory information\n");
     num = len - 4;
     ucp = &gBuf[0] + 4;
     while (num > 3) {
         pc = (ucp[0] << 8) | ucp[1];
         pl = ucp[3] + 4;
+        good = 0;
         switch (pc) {
-        case 0: pout("  number of hours powered up"); break;
+        case 0: pout("  number of hours powered up");
+            good = 1;
+            break;
         case 8: pout("  number of minutes until next internal SMART test");
+            good = 1;
+            break;
+        default:
+            if (con->reportscsiioctl > 0) {
+                PRINT_ON(con);
+                pout("Vendor (Seagate/Hitachi) factory lpage: "
+                     "unknown parameter code [0x%x]\n", pc);
+                PRINT_OFF(con);
+            }
             break;
-        default: pout("  Unknown parameter code [0x%x]", pc); break;
         }
-        k = pl - 4;
-        xp = ucp + 4;
-        if (k > (int)sizeof(ull)) {
-            xp += (k - (int)sizeof(ull));
-            k = (int)sizeof(ull);
+        if (good) {
+            k = pl - 4;
+            xp = ucp + 4;
+            if (k > (int)sizeof(ull)) {
+                xp += (k - (int)sizeof(ull));
+                k = (int)sizeof(ull);
+            }
+            ull = 0;
+            for (j = 0; j < k; ++j) {
+                if (j > 0)
+                    ull <<= 8;
+                ull |= xp[j];
+            }
+            if (0 == pc)
+                pout(" = %.2f\n", uint64_to_double(ull) / 60.0 );
+            else
+                pout(" = %"PRIu64"\n", ull);
         }
-        ull = 0;
-        for (j = 0; j < k; ++j) {
-            if (j > 0)
-                ull <<= 8;
-            ull |= xp[j];
-        }
-        if (0 == pc)
-            pout(" = %.2f\n", uint64_to_double(ull) / 60.0 );
-        else
-            pout(" = %"PRIu64"\n", ull);
         num -= pl;
         ucp += pl;
     }
@@ -409,18 +503,18 @@ static void scsiPrintErrorCounterLog(int device)
     int k;
     double processed_gb;
 
-    if (0 == scsiLogSense(device, READ_ERROR_COUNTER_LPAGE, gBuf, 
-                          LOG_RESP_LEN, 0)) {
+    if (gReadECounterLPage && (0 == scsiLogSense(device,
+                READ_ERROR_COUNTER_LPAGE, gBuf, LOG_RESP_LEN, 0))) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[0]);
         found[0] = 1;
     }
-    if (0 == scsiLogSense(device, WRITE_ERROR_COUNTER_LPAGE, gBuf, 
-                          LOG_RESP_LEN, 0)) {
+    if (gWriteECounterLPage && (0 == scsiLogSense(device,
+                WRITE_ERROR_COUNTER_LPAGE, gBuf, LOG_RESP_LEN, 0))) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[1]);
         found[1] = 1;
     }
-    if (0 == scsiLogSense(device, VERIFY_ERROR_COUNTER_LPAGE, gBuf, 
-                          LOG_RESP_LEN, 0)) {
+    if (gVerifyECounterLPage && (0 == scsiLogSense(device,
+                VERIFY_ERROR_COUNTER_LPAGE, gBuf, LOG_RESP_LEN, 0))) {
         scsiDecodeErrCounterPage(gBuf, &errCounterArr[2]);
         ecp = &errCounterArr[2];
         for (k = 0; k < 7; ++k) {
@@ -434,7 +528,7 @@ static void scsiPrintErrorCounterLog(int device)
         pout("\nError counter log:\n");
         pout("           Errors Corrected by           Total   "
              "Correction     Gigabytes    Total\n");
-        pout("               EEC          rereads/    errors   "
+        pout("               ECC          rereads/    errors   "
              "algorithm      processed    uncorrected\n");
         pout("           fast | delayed   rewrites  corrected  "
              "invocations   [10^9 bytes]  errors\n");
@@ -451,8 +545,8 @@ static void scsiPrintErrorCounterLog(int device)
     }
     else 
         pout("\nError Counter logging not supported\n");
-    if (0 == scsiLogSense(device, NON_MEDIUM_ERROR_LPAGE, gBuf, 
-                          LOG_RESP_LEN, 0)) {
+    if (gNonMediumELPage && (0 == scsiLogSense(device,
+                NON_MEDIUM_ERROR_LPAGE, gBuf, LOG_RESP_LEN, 0))) {
         scsiDecodeNonMediumErrPage(gBuf, &nme);
         if (nme.gotPC0)
             pout("\nNon-medium error count: %8"PRIu64"\n", nme.counterPC0);
@@ -463,18 +557,19 @@ static void scsiPrintErrorCounterLog(int device)
             pout("Positioning error count [Hitachi]: %8"PRIu64"\n",
                  nme.counterPE_H);
     }
-    if (0 == scsiLogSense(device, LAST_N_ERROR_LPAGE, gBuf, 
-                          LOG_RESP_LONG_LEN, 0)) {
+    if (gLastNErrorLPage && (0 == scsiLogSense(device,
+                LAST_N_ERROR_LPAGE, gBuf, LOG_RESP_LONG_LEN, 0))) {
         unsigned char * ucp;
         int num, k, pc, pl;
 
-        num = (gBuf[2] << 8) + gBuf[3];
+        num = (gBuf[2] << 8) + gBuf[3] + 4;
         num = (num < LOG_RESP_LONG_LEN) ? num : LOG_RESP_LONG_LEN;
         ucp = gBuf + 4;
+        num -= 4;
         if (num < 4)
             pout("\nNo error events logged\n");
         else {
-            pout("Last n error events log page\n");
+            pout("\nLast n error events log page\n");
             for (k = num; k > 0; k -= pl, ucp += pl) {
                 if (k < 3) {
                     pout("  <<short Last n error events log page>>\n");
@@ -482,21 +577,28 @@ static void scsiPrintErrorCounterLog(int device)
                 }
                 pl = ucp[3] + 4;
                 pc = (ucp[0] << 8) + ucp[1];
-                pout("  Error event %d:\n", pc);
-                if (ucp[2] & 0x2) {
-                    pout("    [binary]:\n");
-                    dStrHex((const char *)ucp + 4, pl - 4, 1);
-                } else
-                    pout("    %.*s\n", pl - 4, (const char *)(ucp + 4));
-                num -= pl;
-                ucp += pl;
+                if (pl > 4) {
+                    if ((ucp[2] & 0x1) && (ucp[2] & 0x2)) {
+                        pout("  Error event %d:\n", pc);
+                        pout("    [binary]:\n");
+                        dStrHex((const char *)ucp + 4, pl - 4, 1);
+                    } else if (ucp[2] & 0x1) {
+                        pout("  Error event %d:\n", pc);
+                        pout("    %.*s\n", pl - 4, (const char *)(ucp + 4));
+                    } else {
+                        if (con->reportscsiioctl > 0) {
+                            pout("  Error event %d:\n", pc);
+                            pout("    [data counter??]:\n");
+                            dStrHex((const char *)ucp + 4, pl - 4, 1);
+			}
+                    }
+                }
             }
         }
-    } else
-        pout("\nError Events logging not supported\n");
+    }
 }
 
-const char * self_test_code[] = {
+static const char * self_test_code[] = {
         "Default         ", 
         "Background short", 
         "Background long ", 
@@ -507,7 +609,7 @@ const char * self_test_code[] = {
         "Reserved(7)     "
 };
 
-const char * self_test_result[] = {
+static const char * self_test_result[] = {
         "Completed                ",
         "Interrupted ('-X' switch)",
         "Interrupted (bus reset ?)",
@@ -526,12 +628,15 @@ const char * self_test_result[] = {
         "Self test in progress ..."
 };
 
-// See Working Draft SCSI Primary Commands - 3 (SPC-3) pages 231-233
-// T10/1416-D Rev 10
+// See SCSI Primary Commands - 3 (SPC-3) rev 23 (draft) section 7.2.10 .
+// Returns 0 if ok else FAIL* bitmask. Note that if any of the most recent
+// 20 self tests fail (result code 3 to 7 inclusive) then FAILLOG and/or
+// FAILSMART is returned.
 static int scsiPrintSelfTest(int device)
 {
     int num, k, n, res, err, durationSec;
     int noheader = 1;
+    int retval = 0;
     UINT8 * ucp;
     uint64_t ull=0;
 
@@ -540,13 +645,13 @@ static int scsiPrintSelfTest(int device)
         PRINT_ON(con);
         pout("scsiPrintSelfTest Failed [%s]\n", scsiErrString(err));
         PRINT_OFF(con);
-        return 1;
+        return FAILSMART;
     }
     if (gBuf[0] != SELFTEST_RESULTS_LPAGE) {
         PRINT_ON(con);
         pout("Self-test Log Sense Failed, page mismatch\n");
         PRINT_OFF(con);
-        return 1;
+        return FAILSMART;
     }
     // compute page length
     num = (gBuf[2] << 8) + gBuf[3];
@@ -555,7 +660,7 @@ static int scsiPrintSelfTest(int device)
         PRINT_ON(con);
         pout("Self-test Log Sense length is 0x%x not 0x190 bytes\n",num);
         PRINT_OFF(con);
-        return 1;
+        return FAILSMART;
     }
     // loop through the twenty possible entries
     for (k = 0, ucp = gBuf + 4; k < 20; ++k, ucp += 20 ) {
@@ -583,8 +688,38 @@ static int scsiPrintSelfTest(int device)
         pout("#%2d  %s", (ucp[0] << 8) | ucp[1], 
             self_test_code[(ucp[4] >> 5) & 0x7]);
 
-        // self-test result
-        res = ucp[4] & 0xf;
+        // check the self-test result nibble, using the self-test results
+        // field table from T10/1416-D (SPC-3) Rev. 23, section 7.2.10:
+        switch ((res = ucp[4] & 0xf)) {
+        case 0x3:
+            // an unknown error occurred while the device server
+            // was processing the self-test and the device server
+            // was unable to complete the self-test
+            retval|=FAILSMART;
+            break;
+        case 0x4:
+            // the self-test completed with a failure in a test
+            // segment, and the test segment that failed is not
+            // known
+            retval|=FAILLOG;
+            break;
+        case 0x5:
+            // the first segment of the self-test failed
+            retval|=FAILLOG;
+            break;
+        case 0x6:
+            // the second segment of the self-test failed
+            retval|=FAILLOG;
+            break;
+        case 0x7:
+            // another segment of the self-test failed and which
+            // test is indicated by the contents of the SELF-TEST
+            // NUMBER field
+            retval|=FAILLOG;
+            break;
+        default:
+            break;
+        }
         pout("  %s", self_test_result[res]);
 
         // self-test number identifies test that failed and consists
@@ -601,9 +736,9 @@ static int scsiPrintSelfTest(int device)
         // print time that the self-test was completed
         if (n==0 && res==0xf)
         // self-test in progress
-            pout("   NOW");
+            pout("     NOW");
         else   
-            pout(" %5d", n);
+            pout("   %5d", n);
           
         // construct 8-byte integer address of first failure
         for (i = 0; i < 8; i++) {
@@ -611,10 +746,15 @@ static int scsiPrintSelfTest(int device)
             ull |= ucp[i+8];
         }
         // print Address of First Failure, if sensible
-        if ((~(uint64_t)0 != ull) && (res > 0) && (res < 0xf))
-            pout("  0x%16"PRIx64, ull);
-        else
-            pout("                   -");
+        if ((~(uint64_t)0 != ull) && (res > 0) && (res < 0xf)) {
+            char buff[32];
+
+            // was hex but change to decimal to conform with ATA
+            snprintf(buff, sizeof(buff), "%"PRIu64, ull);
+            // snprintf(buff, sizeof(buff), "0x%"PRIx64, ull);
+            pout("%18s", buff);
+        } else
+            pout("                 -");
 
         // if sense key nonzero, then print it, along with
         // additional sense code and additional sense code qualifier
@@ -634,7 +774,7 @@ static int scsiPrintSelfTest(int device)
         pout("Long (extended) Self Test duration: %d seconds "
              "[%.1f minutes]\n", durationSec, durationSec / 60.0);
     }
-    return 0;
+    return retval;
 }
 
 static const char * peripheral_dt_arr[] = {
@@ -729,10 +869,12 @@ static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
     strncpy(revision, (char *)&gBuf[32], 4);
     pout("Device: %s %s Version: %s\n", manufacturer, product, revision);
 
-    if (0 == strncmp(manufacturer, "3ware", 5)) {
-        pout("please try '-d 3ware,N'\n");
+    if (0 == strncmp(manufacturer, "3ware", 5) || 0 == strncmp(manufacturer, "AMCC", 4) ) {
+        pout("please try adding '-d 3ware,N'\n");
+        pout("you may also need to change device to /dev/twaN or /dev/tweN\n");
         return 2;
-    } else if ((len >= 42) && (0 == strncmp(&gBuf[36], "MVSATA", 6))) {
+    } else if ((len >= 42) &&
+	       (0 == strncmp((const char *)(gBuf + 36), "MVSATA", 6))) {
         pout("please try '-d marvell'\n");
         return 2;
     } else if ((avail_len >= 96) && (0 == strncmp(manufacturer, "ATA", 3))) {
@@ -751,10 +893,10 @@ static int scsiGetDriveInfo(int device, UINT8 * peripheral_type, int all)
         avail_len = ((gBuf[2] << 8) + gBuf[3]) + 4;
         len = (avail_len < req_len) ? avail_len : req_len;
         if (isLinuxLibAta(gBuf, len)) {
-            pout("\nSATA disks accessed via libata are not currently "
-                 "supported by\nsmartmontools. When libata is given "
-                 "an ATA pass-thru ioctl() then an\nadditional '-d libata'"
-                 " device type will be added to smartmontools.\n");
+            pout("\nIn Linux, SATA disks accessed via libata are "
+                 "only supported by smartmontools\n"
+                 "for kernel versions 2.6.15 and above. Try "
+                 "an additional '-d ata' argument.\n");
             return 2;
         }
     }
@@ -927,7 +1069,7 @@ static int scsiSmartDisable(int device)
     return 0;
 }
 
-void scsiPrintTemp(int device)
+static void scsiPrintTemp(int device)
 {
     UINT8 temp = 0;
     UINT8 trip = 0;
@@ -999,7 +1141,12 @@ int scsiPrintMain(int fd)
             else
                 pout("TapeAlert Not Supported\n");
         } else { /* disk, cd/dvd, enclosure, etc */
-            scsiGetSmartData(fd, con->smartvendorattrib);
+            if ((res = scsiGetSmartData(fd, con->smartvendorattrib))) {
+                if (-2 == res)
+                    returnval |= FAILSTATUS;
+                else
+                    returnval |= FAILSMART;
+            }
         }
     }   
     if (con->smartvendorattrib) {
@@ -1013,6 +1160,7 @@ int scsiPrintMain(int fd)
         if (gStartStopLPage)
             scsiGetStartStopData(fd);
         if (SCSI_PT_DIRECT_ACCESS == peripheral_type) {
+            scsiPrintGrownDefectListLen(fd);
             if (gSeagateCacheLPage)
                 scsiPrintSeagateCacheLPage(fd);
             if (gSeagateFactoryLPage)
@@ -1020,6 +1168,8 @@ int scsiPrintMain(int fd)
         }
     }
     if (con->smarterrorlog) {
+        if (! checkedSupportedLogPages)
+            scsiGetSupportedLogPages(fd);
         scsiPrintErrorCounterLog(fd);
         if (1 == scsiFetchControlGLTSD(fd, modese_len, 1))
             pout("\n[GLTSD (Global Logging Target Save Disable) set. "
@@ -1036,35 +1186,27 @@ int scsiPrintMain(int fd)
             failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
         }
         if (0 != res)
-            failuretest(OPTIONAL_CMD, returnval|=FAILSMART);
+            failuretest(OPTIONAL_CMD, returnval|=res);
     }
     if (con->smartexeoffimmediate) {
-        if (scsiSmartDefaultSelfTest(fd)) {
-            pout( "Default Self Test Failed\n");
-            return returnval;
-        }
+        if (scsiSmartDefaultSelfTest(fd))
+            return returnval | FAILSMART;
         pout("Default Self Test Successful\n");
     }
     if (con->smartshortcapselftest) {
-        if (scsiSmartShortCapSelfTest(fd)) {
-            pout("Short Foreground Self Test Failed\n");
-            return returnval;
-        }
+        if (scsiSmartShortCapSelfTest(fd))
+            return returnval | FAILSMART;
         pout("Short Foreground Self Test Successful\n");
     }
     if (con->smartshortselftest ) { 
-        if ( scsiSmartShortSelfTest(fd)) {
-            pout("Short Background Self Test Failed\n");
-            return returnval;
-        }
+        if (scsiSmartShortSelfTest(fd))
+            return returnval | FAILSMART;
         pout("Short Background Self Test has begun\n");
         pout("Use smartctl -X to abort test\n");
     }
     if (con->smartextendselftest) {
-        if (scsiSmartExtendSelfTest(fd)) {
-            pout("Extended Background Self Test Failed\n");
-            return returnval;
-        }
+        if (scsiSmartExtendSelfTest(fd))
+            return returnval | FAILSMART;
         pout("Extended Background Self Test has begun\n");
         if ((0 == scsiFetchExtendedSelfTestTime(fd, &durationSec, 
                         modese_len)) && (durationSec > 0)) {
@@ -1078,17 +1220,13 @@ int scsiPrintMain(int fd)
         pout("Use smartctl -X to abort test\n");        
     }
     if (con->smartextendcapselftest) {
-        if (scsiSmartExtendCapSelfTest(fd)) {
-            pout("Extended Foreground Self Test Failed\n");
-            return returnval;
-        }
+        if (scsiSmartExtendCapSelfTest(fd))
+            return returnval | FAILSMART;
         pout("Extended Foreground Self Test Successful\n");
     }
     if (con->smartselftestabort) {
-        if (scsiSmartSelfTestAbort(fd)) {
-            pout("Self Test Abort Failed\n");
-            return returnval;
-        }
+        if (scsiSmartSelfTestAbort(fd))
+            return returnval | FAILSMART;
         pout("Self Test returned without error\n");
     }           
     return returnval;
