@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -22,22 +22,24 @@
  *
  */
 
-#include "config.h"
 #include <errno.h>
 #include <stdio.h>
 #include <sys/types.h>
 #include <string.h>
 #include <stdarg.h>
+
+#include "config.h"
 #ifdef HAVE_GETOPT_LONG
 #include <getopt.h>
 #endif
 #if defined(__FreeBSD_version) && (__FreeBSD_version < 500000)
 #include <unistd.h>
 #endif
+
+#include "int64.h"
 #include "atacmds.h"
 #include "ataprint.h"
 #include "extern.h"
-#include "int64.h"
 #include "knowndrives.h"
 #include "scsicmds.h"
 #include "scsiprint.h"
@@ -47,11 +49,8 @@
 #ifdef NEED_SOLARIS_ATA_CODE
 extern const char *os_solaris_ata_s_cvsid;
 #endif
-#if defined(_WIN32) && defined(_MSC_VER)
-extern const char *int64_vc6_c_cvsid;
-#endif
 extern const char *atacmdnames_c_cvsid, *atacmds_c_cvsid, *ataprint_c_cvsid, *knowndrives_c_cvsid, *os_XXXX_c_cvsid, *scsicmds_c_cvsid, *scsiprint_c_cvsid, *utility_c_cvsid;
-const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.135 2004/08/18 19:27:45 likewise Exp $"
+const char* smartctl_c_cvsid="$Id: smartctl.c,v 1.143 2006/04/12 14:54:28 ballen4705 Exp $"
 ATACMDS_H_CVSID ATAPRINT_H_CVSID CONFIG_H_CVSID EXTERN_H_CVSID INT64_H_CVSID KNOWNDRIVES_H_CVSID SCSICMDS_H_CVSID SCSIPRINT_H_CVSID SMARTCTL_H_CVSID UTILITY_H_CVSID;
 
 // This is a block containing all the "control variables".  We declare
@@ -65,7 +64,12 @@ extern int exitstatus;
 extern int64_t bytes;
 
 void printslogan(){
-  pout("smartctl version %s [%s] Copyright (C) 2002-4 Bruce Allen\n", PACKAGE_VERSION, SMARTMONTOOLS_BUILD_HOST);
+#ifdef HAVE_GET_OS_VERSION_STR
+  const char * ver = get_os_version_str();
+#else
+  const char * ver = SMARTMONTOOLS_BUILD_HOST;
+#endif
+  pout("smartctl version %s [%s] Copyright (C) 2002-6 Bruce Allen\n", PACKAGE_VERSION, ver);
   pout("Home page is " PACKAGE_HOMEPAGE "\n\n");
   return;
 }
@@ -88,9 +92,6 @@ void printcopy(){
   PrintOneCVS(atacmdnames_c_cvsid);
   PrintOneCVS(atacmds_c_cvsid);
   PrintOneCVS(ataprint_c_cvsid);
-#if defined(_WIN32) && defined(_MSC_VER)
-  PrintOneCVS(int64_vc6_c_cvsid);
-#endif
   PrintOneCVS(knowndrives_c_cvsid);
   PrintOneCVS(os_XXXX_c_cvsid);
 #ifdef NEED_SOLARIS_ATA_CODE
@@ -233,7 +234,7 @@ const char *getvalidarglist(char opt) {
   case 'q':
     return "errorsonly, silent";
   case 'd':
-    return "ata, scsi, 3ware,N";
+    return "ata, scsi, marvell, 3ware,N";
   case 'T':
     return "normal, conservative, permissive, verypermissive";
   case 'b':
@@ -563,7 +564,12 @@ void ParseOpts (int argc, char** argv){
       } else if (!strcmp(optarg, "show")) {
         con->showpresets = TRUE;
       } else if (!strcmp(optarg, "showall")) {
-        showallpresets();
+        if (optind < argc) { // -P showall MODEL [FIRMWARE]
+          int cnt = showmatchingpresets(argv[optind], (optind+1<argc ? argv[optind+1] : NULL));
+          EXIT(cnt); // report #matches
+        }
+        if (showallpresets())
+          EXIT(FAILCMD); // report regexp syntax error
         EXIT(0);
       } else {
         badarg = TRUE;
