@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -35,13 +35,16 @@
 #include <syslog.h>
 #include <stdarg.h>
 #include <sys/stat.h>
+#ifdef _WIN32
+#include <mbstring.h> // _mbsinc()
+#endif
 
 #include "config.h"
 #include "int64.h"
 #include "utility.h"
 
 // Any local header files should be represented by a CVSIDX just below.
-const char* utility_c_cvsid="$Id: utility.c,v 1.54 2004/08/29 01:08:37 ballen4705 Exp $"
+const char* utility_c_cvsid="$Id: utility.c,v 1.61 2006/04/12 14:54:28 ballen4705 Exp $"
 CONFIG_H_CVSID INT64_H_CVSID UTILITY_H_CVSID;
 
 const char * packet_types[] = {
@@ -174,6 +177,30 @@ void FixGlibcTimeZoneBug(){
   return;
 }
 
+#ifdef _WIN32
+// Fix strings in tzname[] to avoid long names with non-ascii characters.
+// If TZ is not set, tzset() in the MSVC runtime sets tzname[] to the
+// national language timezone names returned by GetTimezoneInformation().
+static char * fixtzname(char * dest, int destsize, const char * src)
+{
+  int i = 0, j = 0;
+  while (src[i] && j < destsize-1) {
+    int i2 = (const char *)_mbsinc((const unsigned char *)src+i) - src;
+    if (i2 > i+1)
+      i = i2; // Ignore multibyte chars
+    else {
+      if ('A' <= src[i] && src[i] <= 'Z')
+        dest[j++] = src[i]; // "Pacific Standard Time" => "PST"
+      i++;
+    }
+  }
+  if (j < 2)
+    j = 0;
+  dest[j] = 0;
+  return dest;
+}
+#endif // _WIN32
+
 // This value follows the peripheral device type value as defined in
 // SCSI Primary Commands, ANSI INCITS 301:1997.  It is also used in
 // the ATA standard for packet devices to define the device type.
@@ -206,6 +233,9 @@ void dateandtimezoneepoch(char *buffer, time_t tval){
   char *timezonename;
   char datebuffer[DATEANDEPOCHLEN];
   int lenm1;
+#ifdef _WIN32
+  char tzfixbuf[6+1];
+#endif
 
   FixGlibcTimeZoneBug();
   
@@ -232,6 +262,12 @@ void dateandtimezoneepoch(char *buffer, time_t tval){
   else
     // unable to determine if daylight savings in effect
     timezonename="";
+
+#ifdef _WIN32
+  // Fix long non-ascii timezone names
+  if (!getenv("TZ"))
+    timezonename=fixtzname(tzfixbuf, sizeof(tzfixbuf), timezonename);
+#endif
   
   // Finally put the information into the buffer as needed.
   snprintf(buffer, DATEANDEPOCHLEN, "%s %s", datebuffer, timezonename);
@@ -260,38 +296,34 @@ int massagecvs(char *out, const char *cvsid){
   if (!(copy=strdup(cvsid)))
     return 0;
 
-  if (!(filename=strtok(copy, delimiters))){
-    free(copy);
-    return 0;
-  }
+  if (!(filename=strtok(copy, delimiters)))
+    goto endmassage;
 
   // move to first instance of "Id:"
   while (strcmp(filename,"Id:"))
-    if (!(filename=strtok(NULL, delimiters))){
-      free(copy);
-      return 0;
-    }
+    if (!(filename=strtok(NULL, delimiters)))
+      goto endmassage;
   
   // get filename, skip "v", get version and date
   if (!(  filename=strtok(NULL, delimiters)  ) ||
       !(           strtok(NULL, delimiters)  ) ||
       !(   version=strtok(NULL, delimiters)  ) ||
-      !(      date=strtok(NULL, delimiters)  ) ) {
-    free(copy);
-    return 0;
-  }
+      !(      date=strtok(NULL, delimiters)  ) )
+    goto endmassage;
   
   sprintf(out,"%-16s revision: %-5s date: %-15s", filename, version, date);
   retVal = (date-copy)+strlen(date);
+  
+ endmassage:
   free(copy);
-  return  retVal;
+  return retVal;
 }
 
 // prints a single set of CVS ids
 void printone(char *block, const char *cvsid){
   char strings[CVSMAXLEN];
   const char *here=cvsid;
-  int line=1,len=strlen(cvsid)+1;
+  int bi=0, len=strlen(cvsid)+1;
 
   // check that the size of the output block is sufficient
   if (len>=CVSMAXLEN) {
@@ -300,15 +332,8 @@ void printone(char *block, const char *cvsid){
   }
 
   // loop through the different strings
-  while ((len=massagecvs(strings,here))){
-    switch (line++){
-    case 1:
-      block+=snprintf(block,CVSMAXLEN,"Module:");
-      break;
-    default:
-      block+=snprintf(block,CVSMAXLEN,"  uses:");
-    } 
-    block+=snprintf(block,CVSMAXLEN," %s\n",strings);
+  while (bi<CVSMAXLEN && (len=massagecvs(strings,here))){
+    bi+=snprintf(block+bi,CVSMAXLEN-bi,"%s %s\n",(bi==0?"Module:":"  uses:"),strings);
     here+=len;
   }
   return;
@@ -351,14 +376,40 @@ void printregexwarning(int errcode, regex_t *compiled){
   return;
 }
 
+// POSIX extended regular expressions interpret unmatched ')' ordinary:
+// "The close-parenthesis shall be considered special in this context
+//  only if matched with a preceding open-parenthesis."
+//
+// Actual '(...)' nesting errors remain undetected on strict POSIX
+// implementations (glibc) but an error is reported on others (Cygwin).
+// 
+// The check below is rather incomplete because it does not handle
+// e.g. '\)' '[)]'.
+// But it should work for the regex subset used in drive database.
+static int check_regex_nesting(const char * pattern)
+{
+  int level = 0, i;
+  for (i = 0; pattern[i] && level >= 0; i++) {
+    switch (pattern[i]) {
+      case '(': level++; break;
+      case ')': level--; break;
+    }
+  }
+  return level;
+}
+
 // A wrapper for regcomp().  Returns zero for success, non-zero otherwise.
 int compileregex(regex_t *compiled, const char *pattern, int cflags)
 { 
   int errorcode;
 
-  if ((errorcode = regcomp(compiled, pattern, cflags))) {
-    pout("Internal error: unable to compile regular expression %s", pattern);
-    printregexwarning(errorcode, compiled);
+  if (   (errorcode = regcomp(compiled, pattern, cflags))
+      || check_regex_nesting(pattern) < 0                ) {
+    pout("Internal error: unable to compile regular expression \"%s\" ", pattern);
+    if (errorcode)
+      printregexwarning(errorcode, compiled);
+    else
+      pout("Unmatched ')'\n");
     pout("Please inform smartmontools developers at " PACKAGE_BUGREPORT "\n");
     return 1;
   }
@@ -596,4 +647,35 @@ void MsecToText(unsigned int msec, char *txt){
 }
 
 
+#ifndef HAVE_WORKING_SNPRINTF
+// Some versions of (v)snprintf() don't append null char on overflow (MSVCRT.DLL),
+// and/or return -1 on overflow (old Linux).
+// Below are sane replacements substituted by #define in utility.h.
 
+#undef vsnprintf
+#if defined(_WIN32) && defined(_MSC_VER)
+#define vsnprintf _vsnprintf
+#endif
+
+int safe_vsnprintf(char *buf, int size, const char *fmt, va_list ap)
+{
+  int i;
+  if (size <= 0)
+    return 0;
+  i = vsnprintf(buf, size, fmt, ap);
+  if (0 <= i && i < size)
+    return i;
+  buf[size-1] = 0;
+  return strlen(buf); // Note: cannot detect for overflow, not necessary here.
+}
+
+int safe_snprintf(char *buf, int size, const char *fmt, ...)
+{
+  int i; va_list ap;
+  va_start(ap, fmt);
+  i = safe_vsnprintf(buf, size, fmt, ap);
+  va_end(ap);
+  return i;
+}
+
+#endif
