@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2002-4 Bruce Allen <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2002-6 Bruce Allen <smartmontools-support@lists.sourceforge.net>
  * Copyright (C) 2000 Michael Cornwell <cornwell@acm.org>
  *
  * This program is free software; you can redistribute it and/or modify
@@ -25,17 +25,19 @@
 #ifndef UTILITY_H_
 #define UTILITY_H_
 
-#define UTILITY_H_CVSID "$Id: utility.h,v 1.38 2004/08/18 19:27:36 likewise Exp $\n"
+#define UTILITY_H_CVSID "$Id: utility.h,v 1.43 2006/04/12 14:54:28 ballen4705 Exp $\n"
 
 #include <time.h>
 #include <sys/types.h> // for regex.h (according to POSIX)
 #include <regex.h>
 
-#include "int64.h"
-
-#if defined(_WIN32) && defined(_MSC_VER)
-#define snprintf  _snprintf
-#define vsnprintf  _vsnprintf
+#ifndef HAVE_WORKING_SNPRINTF
+// Substitute by safe replacement functions
+#include <stdarg.h>
+int safe_snprintf(char *buf, int size, const char *fmt, ...);
+int safe_vsnprintf(char *buf, int size, const char *fmt, va_list ap);
+#define snprintf  safe_snprintf
+#define vsnprintf safe_vsnprintf
 #endif
 
 // Utility function prints current date and time and timezone into a
@@ -127,6 +129,12 @@ int deviceopen(const char *pathname, char *type);
 
 int deviceclose(int fd);
 
+// Optional functions of os_*.c
+#ifdef HAVE_GET_OS_VERSION_STR
+// Return build host and OS version as static string
+const char * get_os_version_str(void);
+#endif
+
 // returns 1 if any of the n bytes are nonzero, else zero.
 int nonempty(unsigned char *testarea,int n);
 
